;;  Machine description of the Axis CRIS architecture (Etrax, Etrax 100).
;;  Copyright (C) 1992 .. 1998, 1999 Free Software Foundation, Inc.

;;  Contributed by Hans-Peter Nilsson <hp@axis.se>, Axis Communications AB.

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;
;; Instruction patterns.  When multiple patterns apply,
;; the first defined one is chosen.
;;
;; See files "md.texi" and "rtl.def" for documentation on define_insn,
;; match_*, et. al.
;;
;; The function "notice_update_cc ()" in cris.c handles condition code
;; updates for most instructions.

;; There are several instructions that are orthogonal in size, and seems
;; they could be matched by a single pattern without a specified size
;; for the operand that is orthogonal.	However, this did not work on
;; gcc-2.7.2 (and problably not on gcc-2.8.1), relating to that when a
;; constant is substituted, the actual mode must be deduced from the
;; pattern.  (There is reasonable hope that that has been fixed in egcs
;; post 1.1.1, so try it then with a huge regression-test-suite).
;;  Maybe I went overboard when I heard about not being able to omit the
;; mode (from Kenner); I "expanded" all patterns with brains off (to
;; keep it safe :-)

;; You will notice that three-operand alternatives ("=r", "r", "!To")
;; are marked with the exclamation mark to be avoided for reloading
;; into.  This is because gcc would otherwise prefer to use the constant
;; pool and its offsettable address instead of reloading to an
;; ("=r", "0", "i") alternative.  (And the constant pool support was not
;; only suboptimal but also buggy in 2.7.2 or maybe it was 2.6.3)

;; All insns that look like (set (...) (plus (...) (reg:SI 8)))
;; get problems when reloading r8 (frame pointer) to r14+offs (stack
;; pointer).  Thus the instructions that get into trouble have specific
;; checks against matching frame_pointer_rtx.
;; ??? But it should be re-checked for gcc > 2.7.2

;; We need an attribute to define whether an instruction can be put in
;; a branch-delay slot or not, and whether it has a delay slot.
;;
;; Branches and return instructions has a delay slot, and cannot
;; themselves be put in a delay slot.  This has changed [for short
;; branches only?] between architecture variants, but the possible win
;; is ad-hoc seen as negligible so a description of that variance has
;; not yet cluttered the machine description.  If you implement it and
;; find/write a real program where the win is more than .01% of the
;; execution time, I'll buy you a case of beer or hit you over the
;; head).
;;
;; Whether an instruction can be put in a delay slot depends on the
;; instruction (all instructions except jumps and branches are allowed)
;; and the addressing mode (must not be prefixed or referring to pc).
;; In short, any slottable instruction must be 16 bit and not refer to,
;; or alter pc.
;;
;; The possible values are "yes", "no" and "branch".  Yes/no means if
;; the insn is slottable or not.  Branch means that the insn is a
;; branch-insn (which is not slottable).
;;
;; The default that is defined here for this attribute is "no", not
;; slottable, so there's no need to worry about it being wrong for
;; non-branch and return instructions; only slightly suboptimal.
;;  The default *could* depend on the kind of insn and the addressing
;; mode, but that would need more attributes and hairier and more error
;; prone code.
;;
;;  The method used is to let the constraint guide whether the insn is
;; slottable or not.  This is the shortest way to win and (hopefully)
;; the most understandable.
;;  There is an extra constraint, 'Q', which recognizes indirect reg,
;; except when reg is pc (not possible at the moment; (pc) is not
;; recognized as (REG 15)).  The constraints 'Q' and '>' together match
;; all possible memory operands that are slottable.
;;  For other operands, you need to check if it has a valid "slottable"
;; quick-immediate operand, where the particular signedness-variation
;; may match the constraints 'I' or 'J'.), and include it in the
;; constraint pattern for the slottable pattern.  Any "r" is slottable
;; (as mentioned, it cannot match PC).

(define_attr "slottable" "no,yes,branch" (const_string "no"))

;;
;; We also need attributes to sanely determine the condition code
;; state.  To probe the insn with just "notice_update_cc ()" is
;; madness.
;;  At the present, we make a transition to just using attribute cc, and
;; the default value of indeterminate makes us use the old
;; "notice_update_cc ()" spaghetti.
;;

(define_attr "cc" "none,clobber,normal" (const_string "normal"))

;; A branch is delayed one cycle; the instruction just after the
;; branch is read and executed.	 That instruction is always executed,
;; independent of whether the branch is taken or not.

(define_delay (eq_attr "slottable" "branch")
  [(eq_attr "slottable" "yes") (nil) (nil)])

;; NOTE: besides the attributes, there also has to be a %# at the end
;; of a "delayed" instruction output pattern
;; (see "jump" => "ba %l0%#"), so print_operand can catch it and print
;; a "nop" if necessary.  (This method is snatched from sparc.md)
;;

;;
;; Test insns.
;;

;;
;; DI-mode
;;
;; Allow register and offsettable mem operands only; post-increment is
;; not worth the trouble.
;;
(define_insn "tstdi"
  [(set (cc0)
	(match_operand:DI 0 "nonimmediate_operand" "r,o"))]
  ""
  "test.d %M0\;ax\;test.d %H0")

;; No test insns with side-effect on the mem addressing.
;;
;; See note on cmp-insns with side-effects (or lack of them)

;;
;; Normal named test patterns from SI on.
;;
(define_insn "tstsi"
  [(set (cc0)
	(match_operand:SI 0 "nonimmediate_operand" "r,Q>,m"))]
  ""
  "test.d %0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn "tsthi"
  [(set (cc0)
	(match_operand:HI 0 "nonimmediate_operand" "r,Q>,m"))]
  ""
  "test.w %0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn "tstqi"
  [(set (cc0)
	(match_operand:QI 0 "nonimmediate_operand" "r,Q>,m"))]
  ""
  "test.b %0"
  [(set_attr "slottable" "yes,yes,no")])

;; It seems that the position of the sign-bit and the fact that 0.0 is
;; all 0-bits would make "tstsf" a straight-forward implementation;
;; either "test.d" it for positive/negative or "btstq 30,r" it for
;; zeroness.
;;
;; Maybe do this in the future: check "next_cc0_user ()" to determine if
;; zero or negative is used (or both, then take special precautions).

;;
;; Compare insns.
;;

;;
;; We could optimize the sizes of the immediate operands for various
;; cases, but that is not worth it because of the very little usage of
;; DI-mode as anything else but a structure/block-mode.	 Just do the
;; obvious stuff for the defined constraint letters.
;;
(define_insn "cmpdi"
  [(set (cc0)
	(compare (match_operand:DI 0 "nonimmediate_operand" "r,r,r,r,r,r,o")
		 (match_operand:DI 1 "general_operand" "K,I,P,n,r,o,r")))]
  ""
  "@
   cmpq %1,%M0\;ax\;cmpq 0,%H0
   cmpq %1,%M0\;ax\;cmpq -1,%H0
   cmp%e1.%z1 %1,%M0\;ax\;cmpq %H1,%H0
   cmp.d %M1,%M0\;ax\;cmp.d %H1,%H0
   cmp.d %M1,%M0\;ax\;cmp.d %H1,%H0
   cmp.d %M1,%M0\;ax\;cmp.d %H1,%H0
   cmp.d %M0,%M1\;ax\;cmp.d %H0,%H1")

;; Compare insns with side effect addressing mode (e.g.):
;;
;; cmp.S [rx=ry+i],rz;
;; cmp.S [%3=%1+%2],%0
;;
;; are *not* possible since the reloader DOES NOT ACCEPT cc0-changing
;; insns with side-effects other than setting the condition codes.  The
;; reason is that the reload stage MAY cause another insn to be output
;; after the main instruction, causing the condition codes to be
;; invalid.  (This does not apply to the CRIS case, since move to
;; memory does not change the condition code, but we have no way to
;; force our will with this at the moment.  I think it could help in the
;; order of a percent).

;; We have cmps and cmpu (compare reg w. sign/zero extended mem).
;; Mostly useful to compare in SI-mode with 8/16-bit constants ([pc+]),
;; but sometimes gcc will find its way to use it for other (memory) operands.
;; Avoid side-effect patterns, though (see above).
;;
;; QI-mode
(define_insn ""
  [(set (cc0)
	(compare
	 (match_operand:SI 0 "register_operand" "r,r")
	 (match_operator:SI 2 "extend_operator"
			 [(match_operand:QI 1 "memory_operand" "Q>,m")])))]
  ""
  "cmp%e2.%s1 %1,%0"
  [(set_attr "slottable" "yes,no")])

;; HI-mode
(define_insn ""
  [(set (cc0)
	(compare
	 (match_operand:SI 0 "register_operand" "r,r")
	 (match_operator:SI 2 "extend_operator"
			 [(match_operand:HI 1 "memory_operand" "Q>,m")])))]
  ""
  "cmp%e2.%s1 %1,%0"
  [(set_attr "slottable" "yes,no")])

;;
;; Swap operands; it seems the canonical look (if any) is not enforced.
;;
;; ??? Investigate this some day-with-lower-priorities.
;;
;; QI-mode
(define_insn ""
  [(set (cc0)
	(compare
	 (match_operator:SI 2 "extend_operator"
			    [(match_operand:QI 0 "memory_operand" "Q>,m")])
	 (match_operand:SI 1 "register_operand" "r,r")))]
  ""
  "cmp%e2.%s0 %0,%1" ; The function "notice_update_cc ()" knows about
		     ; swapped operands to compares.
  [(set_attr "slottable" "yes,no")])

;; HI-mode
(define_insn ""
  [(set (cc0)
	(compare
	 (match_operator:SI 2 "extend_operator"
			    [(match_operand:HI 0 "memory_operand" "Q>,m")])
	 (match_operand:SI 1 "register_operand" "r,r")))]
  ""
  "cmp%e2.%s0 %0,%1" ; The function "notice_update_cc ()" knows about
		     ; swapped operands to compares.
  [(set_attr "slottable" "yes,no")])

;;
;; The normal compare patterns, from SI on.
;;
(define_insn "cmpsi"
  [(set (cc0)
	(compare
	 (match_operand:SI 0 "nonimmediate_operand" "r,r,r,r,Q>,Q>,r,r,m,m")
	 (match_operand:SI 1 "general_operand" "I,r,Q>,M,M,r,P,g,M,r")))]
  ""
  "@
   cmpq %1,%0
   cmp.d %1,%0
   cmp.d %1,%0
   test.d %0
   test.d %0
   cmp.d %0,%1
   cmp%e1.%z1 %1,%0
   cmp.d %1,%0
   test.d %0
   cmp.d %0,%1"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,yes,no,no,no,no")])

(define_insn "cmphi"
  [(set (cc0)
	(compare (match_operand:HI 0 "nonimmediate_operand" "r,r,Q>,Q>,r,m,m")
		 (match_operand:HI 1 "general_operand" "r,Q>,M,r,g,M,r")))]
  ""
  "@
   cmp.w %1,%0
   cmp.w %1,%0
   test.w %0
   cmp.w %0,%1
   cmp.w %1,%0
   test.w %0
   cmp.w %0,%1"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no,no")])

(define_insn "cmpqi"
  [(set (cc0)
	(compare
	 (match_operand:QI 0 "nonimmediate_operand" "r,r,r,Q>,Q>,r,m,m")
	 (match_operand:QI 1 "general_operand" "r,Q>,M,M,r,g,M,r")))]
  ""
  "@
   cmp.b %1,%0
   cmp.b %1,%0
   test.b %0
   test.b %0
   cmp.b %0,%1
   cmp.b %1,%0
   test.b %0
   cmp.b %0,%1"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,no,no,no")])

;;
;; Pattern that use the btst insn.
;; It is useful for "if (i & val)" constructs, where val is an exact
;; power of 2, or if val+1 is a power of two (i.e. we check for a bunch
;; of zeros starting at bit 0).
;;
;; SI-mode (will cover all other modes since gcc automatically extends subregs)
(define_insn ""
  [(set (cc0)
	(zero_extract
	 (match_operand:SI 0 "nonmemory_operand" "r,r,r,r,r,r,n")
	 (match_operand:SI 1 "const_int_operand" "K,n,K,n,K,n,n")
	 (match_operand:SI 2 "nonmemory_operand" "M,M,K,n,r,r,r")))]
  ;; Either it is a single bit, or consecutive ones starting at 0.
  "GET_CODE (operands[1]) == CONST_INT
   && (operands[1] == const1_rtx || operands[2] == const0_rtx)
   && (REG_S_P (operands[0])
       || (operands[1] == const1_rtx && REG_S_P (operands[2])
	   && GET_CODE (operands[0]) == CONST_INT
	   && exact_log2 (INTVAL (operands[0])) >= 0))"
;;
;; The last "&&" condition above should be caught by some kind of
;; canonicalization in gcc, but we can easily help with it here.
;;  It results from expressions of the type
;; "power_of_2_value & (1 << y)".
;;
;; Since there may be (buggy) tests in C on bits (in constant position)
;; beyond the size of a word, handle that by assuming those bits are 0.
;;
  "@
   btstq (%1-1),%0
   test.d %0
   btstq %2,%0
   clearf nz
   btst %2,%0
   clearf nz
   cmpq %p0,%2"
 [(set_attr "slottable" "yes")])

;;
;; Move insns.
;;

;; No pattern for movdi is needed, gcc can split into movsi:s by
;; itself.  (It may be that it does not do it optimally, but we should
;; not perform cover-ups here).

;;
;; Side-effect patterns for move.S1 [rx=ry+rx.S2],rw
;; and move.S1 [rx=ry+i],rz
;;  Then movs.S1 and movu.S1 for both modes.
;;
;; move.S1 [rx=ry+rz.S],rw avoiding when rx is ry, or rw is rx
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r")
	(mem:QI (plus:SI
		 (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			  (match_operand:SI 2 "const_int_operand" "n,n"))
		 (match_operand:SI 3 "register_operand" "r,r"))))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   move.%s0 [%4=%3+%1%T2],%0")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(mem:HI (plus:SI
		 (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			  (match_operand:SI 2 "const_int_operand" "n,n"))
		 (match_operand:SI 3 "register_operand" "r,r"))))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   move.%s0 [%4=%3+%1%T2],%0")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(mem:SI (plus:SI
		 (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			  (match_operand:SI 2 "const_int_operand" "n,n"))
		 (match_operand:SI 3 "register_operand" "r,r"))))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   move.%s0 [%4=%3+%1%T2],%0")

;;
;; move.S1 [rx=ry+i],rz avoiding when rx is ry, or rz is rx
;; (and i may be a register too)
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r,r")
	(mem:QI (plus:SI (match_operand:SI 1 "bdap_operand" "%r,r,r")
			 (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri"))))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"move.%s0 [%3=%1%S2],%0\";
}")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(mem:HI (plus:SI (match_operand:SI 1 "bdap_operand" "%r,r,r")
			 (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri"))))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"move.%s0 [%3=%1%S2],%0\";
}")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(mem:SI (plus:SI (match_operand:SI 1 "bdap_operand" "%r,r,r")
			 (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri"))))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"move.%s0 [%3=%1%S2],%0\";
}")

;;
;; Other way around; move to memory.
;;
;; move.s rx,[ry=rx+rw.S]
;;
;; QI-mode
(define_insn ""
  [(set (mem:QI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(match_operand:QI 3 "register_operand" "r,r"))
   (set (match_operand:SI 4 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 4, 2, 0, 1, 3)"
  "@
   #
   move.%s3 %3,[%4=%2+%0%T1]")

;; HI-mode
(define_insn ""
  [(set (mem:HI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(match_operand:HI 3 "register_operand" "r,r"))
   (set (match_operand:SI 4 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 4, 2, 0, 1, 3)"
  "@
   #
   move.%s3 %3,[%4=%2+%0%T1]")

;; SI-mode
(define_insn ""
  [(set (mem:SI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(match_operand:SI 3 "register_operand" "r,r"))
   (set (match_operand:SI 4 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 4, 2, 0, 1, 3)"
  "@
   #
   move.%s3 %3,[%4=%2+%0%T1]")

;;
;; move.s rx,[ry=rz+i]
;;
;; QI-mode
(define_insn ""
  [(set (mem:QI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(match_operand:QI 2 "register_operand" "r,r,r"))
   (set (match_operand:SI 3 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 3, 0, 1, -1, 2)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"move.%s2 %2,[%3=%0%S1]\";
}")

;; HI-mode
(define_insn ""
  [(set (mem:HI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(match_operand:HI 2 "register_operand" "r,r,r"))
   (set (match_operand:SI 3 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 3, 0, 1, -1, 2)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"move.%s2 %2,[%3=%0%S1]\";
}")

;; SI-mode
(define_insn ""
  [(set (mem:SI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(match_operand:SI 2 "register_operand" "r,r,r"))
   (set (match_operand:SI 3 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 3, 0, 1, -1, 2)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"move.%s2 %2,[%3=%0%S1]\";
}")

;;
;; Clear patterns with side-effect to memory.  It is hard to share this
;; pattern among modes, so there will be one for each mode.

;;
;;  clear.d [ry=rx+rw.s2]
;;
(define_insn ""
  [(set (mem:SI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(const_int 0))
   (set (match_operand:SI 3 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 3, 2, 0, 1, -1)"
  "@
   #
   clear.d [%3=%2+%0%T1]")

;;
;; clear.d [ry=rz+i]
;;
(define_insn ""
  [(set (mem:SI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(const_int 0))
   (set (match_operand:SI 2 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 2, 0, 1, -1, -1)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"clear.d [%2=%0%S1]\";
}")

;;
;;  clear.w [ry=rx+rw.s2]
;;
(define_insn ""
  [(set (mem:HI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(const_int 0))
   (set (match_operand:SI 3 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 3, 2, 0, 1, -1)"
  "@
   #
   clear.w [%3=%2+%0%T1]")

;;
;; clear.w [ry=rz+i]
;;
(define_insn ""
  [(set (mem:HI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(const_int 0))
   (set (match_operand:SI 2 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 2, 0, 1, -1, -1)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"clear.w [%2=%0%S1]\";
}")

;;
;;  clear.b [ry=rx+rw.s2]
;;
(define_insn ""
  [(set (mem:QI (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	(const_int 0))
   (set (match_operand:SI 3 "register_operand" "=2,r")
	(plus:SI (mult:SI (match_dup 0)
			  (match_dup 1))
		 (match_dup 2)))]
  "side_effect_mode_ok (MULT, operands, 3, 2, 0, 1, -1)"
  "@
   #
   clear.b [%3=%2+%0%T1]")

;;
;; clear.b [ry=rz+i]
;;
(define_insn ""
  [(set (mem:QI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			 (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	(const_int 0))
   (set (match_operand:SI 2 "register_operand" "=0,r,r")
	(plus:SI (match_dup 0)
		 (match_dup 1)))]
  "side_effect_mode_ok (PLUS, operands, 2, 0, 1, -1, -1)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[1]) != CONST_INT
	  || INTVAL (operands[1]) > 127
	  || INTVAL (operands[1]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[1]), 'J')))
    return \"#\";
  return \"clear.b [%2=%0%S1]\";
}")

;;
;; Normal move patterns from SI on.
;;

(define_insn "movsi"
  [(set
    (match_operand:SI 0 "general_operand" "=r,r,r,Q>,r,Q>,g,r,r,g")
    (match_operand:SI 1 "general_operand" "r,Q>,M,M,I,r,M,n,g,r"))]
  ""
  "*
{
  /* Better to have c-switch here; must optimize the size of the move
     insn.  The alternative is to try to find more constraint
     letters...	 */
  switch (which_alternative)
    {
    case 0:
    case 1:
    case 5:
    case 8:
    case 9:
      return \"move.d %1,%0\";

    case 2:
    case 3:
    case 6:
      return \"clear.d %0\";

    /* Constants -32..31 except 0.  */
    case 4:
      return \"moveq %1,%0\";

    /* We can win a little on constants -32768..-33, 32..65535 */
    case 7:
      if (INTVAL (operands[1]) > 0 && INTVAL (operands[1]) < 65536)
	{
	  if (INTVAL (operands[1]) < 256)
	    return \"movu.b %1,%0\";
	  return \"movu.w %1,%0\";
	}
      else if (INTVAL (operands[1]) >= -32768 && INTVAL (operands[1]) < 32768)
	{
	  if (INTVAL (operands[1]) >= -128 && INTVAL (operands[1]) < 128)
	    return \"movs.b %1,%0\";
	  return \"movs.w %1,%0\";
	}
      return \"move.d %1,%0\";

    default:
      return \"BOGUS: %1 -> %0\";
    }
}"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,yes,no,no,no,no")])

;;
;; Extend operations with side-effect from mem to register
;;
;; movs/movu (always from mem to register)
;;
;; [rx=ry+rz.S]
;;
;; QI-mode -> HI-mode
;;
;; ??? Examine the impact of removing this one (gcc should know to
;; truncate from SI) some day-with-lower-priorities.
;;
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(match_operator:HI
	 5 "extend_operator"
	 [(mem:QI (plus:SI
		   (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			    (match_operand:SI 2 "const_int_operand" "n,n"))
		   (match_operand:SI 3 "register_operand" "r,r")))]))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   mov%e5.%m5 [%4=%3+%1%T2],%0")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 5 "extend_operator"
	 [(mem:QI (plus:SI
		   (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			    (match_operand:SI 2 "const_int_operand" "n,n"))
		   (match_operand:SI 3 "register_operand" "r,r")))]))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   mov%e5.%m5 [%4=%3+%1%T2],%0")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 5 "extend_operator"
	 [(mem:HI (plus:SI
		   (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			    (match_operand:SI 2 "const_int_operand" "n,n"))
		   (match_operand:SI 3 "register_operand" "r,r")))]))
   (set (match_operand:SI 4 "register_operand" "=3,r")
	(plus:SI (mult:SI (match_dup 1)
			  (match_dup 2))
		 (match_dup 3)))]
  "side_effect_mode_ok (MULT, operands, 4, 3, 1, 2, 0)"
  "@
   #
   mov%e5.%m5 [%4=%3+%1%T2],%0")

;;
;; [rx=ry+i]
;;
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(match_operator:HI
	 4 "extend_operator"
	 [(mem:QI (plus:SI
		   (match_operand:SI 1 "bdap_operand" "%r,r,r")
		   (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"mov%e4.%m4 [%3=%1%S2],%0\";
}")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 4 "extend_operator"
	 [(mem:QI (plus:SI
		   (match_operand:SI 1 "bdap_operand" "%r,r,r")
		   (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"mov%e4.%m4 [%3=%1%S2],%0\";
}")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 4 "extend_operator"
	 [(mem:HI (plus:SI
		   (match_operand:SI 1 "bdap_operand" "%r,r,r")
		   (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 3 "register_operand" "=1,r,r")
	(plus:SI (match_dup 1)
		 (match_dup 2)))]
  "side_effect_mode_ok (PLUS, operands, 3, 1, 2, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[2]) != CONST_INT
	  || INTVAL (operands[2]) > 127
	  || INTVAL (operands[2]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')))
    return \"#\";
  return \"mov%e4.%m4 [%3=%1%S2],%0\";
}")

(define_insn "movhi"
  [(set
    (match_operand:HI 0 "general_operand" "=r,r,r,Q>,r,Q>,r,r,r,g,g,r")
    (match_operand:HI 1 "general_operand" "r,Q>,M,M,I,r,L,O,n,M,r,g"))]
  ""
  "*
{
  switch (which_alternative)
    {
    case 0:
    case 1:
    case 5:
    case 10:
    case 11:
      return \"move.w %1,%0\";
    case 2:
    case 3:
    case 9:
      return \"clear.w %0\";
    case 4:
      return \"moveq %1,%0\";
    case 6:
    case 8:
      if (INTVAL (operands[1]) < 256 && INTVAL (operands[1]) >= -128)
	{
	  if (INTVAL (operands[1]) > 0)
	    return \"movu.b %1,%0\";
	  return \"movs.b %1,%0\";
	}
      return \"move.w %1,%0\";
    case 7:
      return \"movEq %b1,%0\";
    default:
      return \"BOGUS: %1 -> %0\";
  }
}"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,yes,no,yes,no,no,no,no")
   (set (attr "cc")
	(if_then_else (eq_attr "alternative" "7")
		      (const_string "clobber")
		      (const_string "normal")))])

(define_insn "movstricthi"
  [(set
    (strict_low_part
     (match_operand:HI 0 "general_operand" "+r,r,r,Q>,Q>,g,r,g"))
    (match_operand:HI 1 "general_operand" "r,Q>,M,M,r,M,g,r"))]
  ""
  "@
   move.w %1,%0
   move.w %1,%0
   clear.w %0
   clear.w %0
   move.w %1,%0
   clear.w %0
   move.w %1,%0
   move.w %1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,no,no,no")])

(define_insn "movqi"
  [(set (match_operand:QI 0 "general_operand" "=r,Q>,r,r,Q>,r,g,g,r,r")
	(match_operand:QI 1 "general_operand" "r,r,Q>,M,M,I,M,r,O,g"))]
  ""
  "@
   move.b %1,%0
   move.b %1,%0
   move.b %1,%0
   clear.b %0
   clear.b %0
   moveq %1,%0
   clear.b %0
   move.b %1,%0
   moveq %b1,%0
   move.b %1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,yes,no,no,yes,no")
   (set (attr "cc")
	(if_then_else (eq_attr "alternative" "8")
		      (const_string "clobber")
		      (const_string "normal")))])

(define_insn "movstrictqi"
  [(set (strict_low_part
	 (match_operand:QI 0 "general_operand" "+r,Q>,r,r,Q>,g,g,r"))
	(match_operand:QI 1 "general_operand" "r,r,Q>,M,M,M,r,g"))]
  ""
  "@
   move.b %1,%0
   move.b %1,%0
   move.b %1,%0
   clear.b %0
   clear.b %0
   clear.b %0
   move.b %1,%0
   move.b %1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,no,no,no")])

;;
;; The valid "quick" bit-patterns are, except for 0.0, denormalized
;; values REALLY close to 0, and some NaN:s (I think; their exponent is
;; all ones).
;;  So I will only include 0.0, which is fairly common...
;; It will use clear, so we know ALL types of immediate 0 never change cc.
;;
(define_insn "movsf"
  [(set (match_operand:SF 0 "general_operand" "=r,Q>,r,r,Q>,g,g,r")
	(match_operand:SF 1 "general_operand" "r,r,Q>,G,G,G,r,g"))]
  ""
  "@
   move.d %1,%0
   move.d %1,%0
   move.d %1,%0
   clear.d %0
   clear.d %0
   clear.d %0
   move.d %1,%0
   move.d %1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,no,no,no")])

;;
;; Sign- and zero extension insns.
;;  Those for integer source operand are ordered with the widest source
;; type first.
;;
(define_insn "extendsidi2"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(sign_extend:DI (match_operand:SI 1 "general_operand" "g")))]
  ""
  "move.d %1,%M0\;bpl 0f\;clear.d %H0\;moveq -1,%H0\\n0:")

(define_insn "extendhidi2"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(sign_extend:DI (match_operand:HI 1 "general_operand" "g")))]
  ""
  "movs.w %1,%M0\;bpl 0f\;clear.d %H0\;moveq -1,%H0\\n0:")

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(sign_extend:SI (match_operand:HI 1 "general_operand" "r,Q>,g")))]
  ""
  "movs.w %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn "extendqidi2"
  [(set (match_operand:DI 0 "register_operand" "=r")
	(sign_extend:DI (match_operand:QI 1 "general_operand" "g")))]
  ""
  "movs.b %1,%M0\;bpl 0f\;clear.d %H0\;moveq -1,%H0\\n0:")

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(sign_extend:SI (match_operand:QI 1 "general_operand" "r,Q>,g")))]
  ""
  "movs.b %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

;; To do a byte->word exension, extend to dword, exept that the top half
;; of the register will be clobbered

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(sign_extend:HI (match_operand:QI 1 "general_operand" "r,Q>,g")))]
  ""
  "movs.b %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

;;
;; Zero-extend.	 The DI-mode ones are synthesized by gcc.
;;
(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(zero_extend:SI
	 (match_operand:HI 1 "nonimmediate_operand" "r,Q>,m")))]
  ""
  "movu.w %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(zero_extend:SI
	 (match_operand:QI 1 "nonimmediate_operand" "r,Q>,m")))]
  ""
  "movu.b %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

; Same comment as sign extend byte->word above applies.

(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(zero_extend:HI
	 (match_operand:QI 1 "nonimmediate_operand" "r,Q>,m")))]
  ""
  "movu.b %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

;; All kinds of arithmetic and logical instructions.
;;
;; First anonymous patterns to match addressing modes with
;; side-effects.
;;
;; op.S [rx=ry+I],rz; (add, sub, or, and, bound).
;;
;; [rx=ry+rz.S]
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r")
	(match_operator:QI
	 6 "orthogonal_operator"
	 [(match_operand:QI 1 "register_operand" "0,0")
	  (mem:QI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(match_operator:HI
	 6 "orthogonal_operator"
	 [(match_operand:HI 1 "register_operand" "0,0")
	  (mem:HI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 6 "orthogonal_operator"
	 [(match_operand:SI 1 "register_operand" "0,0")
	  (mem:SI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;;
;; [rx=ry+i] ([%4=%2+%3])
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r,r")
	(match_operator:QI
	 5 "orthogonal_operator"
	 [(match_operand:QI 1 "register_operand" "0,0,0")
	  (mem:QI (plus:SI
		   (match_operand:SI 2 "bdap_operand" "%r,r,r")
		   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(match_operator:HI
	 5 "orthogonal_operator"
	 [(match_operand:HI 1 "register_operand" "0,0,0")
	  (mem:HI (plus:SI
		   (match_operand:SI 2 "bdap_operand" "%r,r,r")
		   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 5 "orthogonal_operator"
	 [(match_operand:SI 1 "register_operand" "0,0,0")
	  (mem:SI (plus:SI
		   (match_operand:SI 2 "bdap_operand" "%r,r,r")
		   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;;
;; To be able to match all cases for commutative operands we may have to
;; have the following pattern for add, or & and.  I do not know really,
;; but it does not break anything.
;;
;; ??? This really ought to be tested some day-with-lower-priorities.
;;
;; op.S [rx=ry+I],rz;
;;
;; [rx=ry+rz.S]
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r")
	(match_operator:QI
	 6 "commutative_orth_op"
	 [(mem:QI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))
	  (match_operand:QI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(match_operator:HI
	 6 "commutative_orth_op"
	 [(mem:HI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))
	  (match_operand:HI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 6 "commutative_orth_op"
	 [(mem:SI (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))
	  (match_operand:SI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6.%s0 [%5=%4+%2%T3],%0")

;;
;; [rx=ry+i] ([%4=%2+%3])
;;
;; QI-mode
(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r,r,r")
	(match_operator:QI
	 5 "commutative_orth_op"
	 [(mem:QI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))
	  (match_operand:QI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;; HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(match_operator:HI
	 5 "commutative_orth_op"
	 [(mem:HI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))
	  (match_operand:HI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;; SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 5 "commutative_orth_op"
	 [(mem:SI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))
	  (match_operand:SI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5.%s0 [%4=%2%S3],%0\";
}")

;;
;; Normal add operations.
;;
(define_insn "adddi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r,&r,&r")
	(plus:DI (match_operand:DI 1 "register_operand" "%0,0,0,0,r")
		 (match_operand:DI 2 "general_operand" "J,N,P,g,!To")))]
  ""
  "@
   addq %2,%M0\;ax\;addq 0,%H0
   subq %n2,%M0\;ax\;subq 0,%H0
   add%e2.%z2 %2,%M0\;ax\;addq 0,%H0
   add.d %M2,%M0\;ax\;add.d %H2,%H0
   add.d %M2,%M1,%M0\;ax\;add.d %H2,%H1,%H0")

(define_insn "addsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r,r,r,r,r")
	(plus:SI
	 (match_operand:SI 1 "register_operand" "%0,0,0,0,0,0,r,r")
	 (match_operand:SI 2 "general_operand" "r,Q>,J,N,n,g,!To,0")))]
;; The last constraints are due to that after reload, the '%' is not
;; honored, and canonicalization doesnt care about keeping the same
;; register as in destination.	This will happen after insn splitting.
;; gcc <= 2.7.2
 ""
 "*
{
  switch (which_alternative)
    {
    case 0:
    case 1:
      return \"add.d %2,%0\";
    case 2:
      return \"addq %2,%0\";
    case 3:
      return \"subq %n2,%0\";
    case 4:
    /* Known value, but not in -63..63.	 Check if addu/subu may be used.  */
      if (INTVAL (operands[2]) > 0)
	{
	  if (INTVAL (operands[2]) < 256)
	    return \"addu.b %2,%0\";
	  if (INTVAL (operands[2]) < 65536)
	    return \"addu.w %2,%0\";
	}
      else
	{
	  if (INTVAL (operands[2]) >= -255)
	    return \"subu.b %n2,%0\";
	  if (INTVAL (operands[2]) >= -65535)
	    return \"subu.w %n2,%0\";
	}
      return \"add.d %2,%0\";
    case 6:
      return \"add.d %2,%1,%0\";
    case 5:
      return \"add.d %2,%0\";
    case 7:
      return \"add.d %1,%0\";
    default:
      return \"BOGUS addsi %2+%1 -> %0\";
    }
}"
 [(set_attr "slottable" "yes,yes,yes,yes,no,no,no,yes")])

(define_insn "addhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r,r,r")
	(plus:HI (match_operand:HI 1 "register_operand" "%0,0,0,0,0,r")
		 (match_operand:HI 2 "general_operand" "r,Q>,J,N,g,!To")))]
  ""
  "@
   add.w %2,%0
   add.w %2,%0
   addq %2,%0
   subq %n2,%0
   add.w %2,%0
   add.w %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "normal,normal,clobber,clobber,normal,normal")])

(define_insn "addqi3"
  [(set (match_operand:QI 0 "register_operand" "=r,r,r,r,r,r,r")
	(plus:QI (match_operand:QI 1 "register_operand" "%0,0,0,0,0,0,r")
		 (match_operand:QI 2 "general_operand" "r,Q>,J,N,O,g,!To")))]
  ""
  "@
   add.b %2,%0
   add.b %2,%0
   addq %2,%0
   subq %n2,%0
   subQ -%b2,%0
   add.b %2,%0
   add.b %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,yes,no,no")
   (set_attr "cc" "normal,normal,clobber,clobber,clobber,normal,normal")])

;;
;; Subtract.
;;
;; Note that because of insn canonicalization these will SELDOM be used
;; with a known constant as an operand.
;;
(define_insn "subdi3"
  [(set (match_operand:DI 0 "register_operand" "=r,r,r,&r,&r")
	(minus:DI (match_operand:DI 1 "register_operand" "0,0,0,0,r")
		  (match_operand:DI 2 "general_operand" "J,N,P,g,!To")))]
  ""
  "@
   subq %2,%M0\;ax\;subq 0,%H0
   addq %n2,%M0\;ax\;addq 0,%H0
   sub%e2.%z2 %2,%M0\;ax\;subq 0,%H0
   sub.d %M2,%M0\;ax\;sub.d %H2,%H0
   sub.d %M2,%M1,%M0\;ax\;sub.d %H2,%H1,%H0")

(define_insn "subsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r,r,r,r,r")
	(minus:SI
	 (match_operand:SI 1 "register_operand" "0,0,0,0,0,0,0,r")
	 (match_operand:SI 2 "general_operand" "r,Q>,J,N,P,n,g,!To")))]
  ""
;; This does not do the optimal: "addu.w 65535,r0" when %2 is negative.
;; But then again, %2 should not be negative.
  "@
   sub.d %2,%0
   sub.d %2,%0
   subq %2,%0
   addq %n2,%0
   sub%e2.%z2 %2,%0
   sub.d %2,%0
   sub.d %2,%0
   sub.d %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no,no,no")])

(define_insn "subhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r,r,r")
	(minus:HI (match_operand:HI 1 "register_operand" "0,0,0,0,0,r")
		  (match_operand:HI 2 "general_operand" "r,Q>,J,N,g,!To")))]
  ""
  "@
   sub.w %2,%0
   sub.w %2,%0
   subq %2,%0
   addq %n2,%0
   sub.w %2,%0
   sub.w %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "normal,normal,clobber,clobber,normal,normal")])

(define_insn "subqi3"
  [(set (match_operand:QI 0 "register_operand" "=r,r,r,r,r,r")
	(minus:QI (match_operand:QI 1 "register_operand" "0,0,0,0,0,r")
		  (match_operand:QI 2 "general_operand" "r,Q>,J,N,g,!To")))]
  ""
  "@
   sub.b %2,%0
   sub.b %2,%0
   subq %2,%0
   addq %2,%0
   sub.b %2,%0
   sub.b %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "normal,normal,clobber,clobber,normal,normal")])

;;
;; CRIS has some add-with-sign-extend and add-with-zero-extend instructions.
;;  Although these perform sign/zero-extension to SImode, they are
;; equally applicable for the HImode case.
;; ??? But maybe should not be described?
;;  Note that these must be located after the normal add/sub patterns,
;; so not to get constants into any unspecified operands.
;;
;; Extend with add/sub and side-effect.
;;
;; adds/subs/addu/subu also bound, check for zero_extend
;;
;; adds/subs/addu/subu bound [rx=ry+rz.S]
;;
;; QI-mode -> HI-mode
;; ??? Maybe remove this; it *should* be redundant to gcc.
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(match_operator:HI
	 6 "operand_extend_operator"
	 [(match_operand:HI 1 "register_operand" "0,0")
	  (match_operator:HI
	   7 "extend_operator"
	   [(mem:QI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[7]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6%e7.%m7 [%5=%4+%2%T3],%0")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 6 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0")
	  (match_operator:SI
	   7 "extend_operator"
	   [(mem:QI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[7]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6%e7.%m7 [%5=%4+%2%T3],%0")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 6 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0")
	  (match_operator:SI
	   7 "extend_operator"
	   [(mem:HI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[7]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x6%e7.%m7 [%5=%4+%2%T3],%0")

;;
;; [rx=ry+i]
;;
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(match_operator:HI
	 5 "operand_extend_operator"
	 [(match_operand:HI 1 "register_operand" "0,0,0")
	  (match_operator:HI
	   6 "extend_operator"
	   [(mem:QI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")
			  ))])]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5%e6.%m6 [%4=%2%S3],%0\";
}")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 5 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0,0")
	  (match_operator:SI
	   6 "extend_operator"
	   [(mem:QI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")
			     ))])]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]

  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5%e6.%m6 [%4=%2%S3],%0\";
}")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 5 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0,0")
	  (match_operator:SI
	   6 "extend_operator"
	   [(mem:HI (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")
			     ))])]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "(GET_CODE (operands[5]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x5%e6.%m6 [%4=%2%S3],%0\";
}")

;;
;; As with op.S we may have to add special pattern to match commuted
;; operands to adds/addu  and bound
;;
;; adds/addu/bound [rx=ry+rz.S]
;;
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(match_operator:HI
	 7 "plus_or_bound_operator"
	 [(match_operator:HI
	   6 "extend_operator"
	   [(mem:QI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])
	  (match_operand:HI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x7%e6.%m6 [%5=%4+%2%T3],%0")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 7 "plus_or_bound_operator"
	 [(match_operator:SI
	   6 "extend_operator"
	   [(mem:QI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])
	  (match_operand:SI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x7%e6.%m6 [%5=%4+%2%T3],%0")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(match_operator:SI
	 7 "plus_or_bound_operator"
	 [(match_operator:SI
	   6 "extend_operator"
	   [(mem:HI (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])
	  (match_operand:SI 1 "register_operand" "0,0")]))
   (set (match_operand:SI 5 "register_operand" "=4,r")
	(plus:SI (mult:SI (match_dup 2)
			  (match_dup 3))
		 (match_dup 4)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[6]) == ZERO_EXTEND)
   && side_effect_mode_ok (MULT, operands, 5, 4, 2, 3, 0)"
  "@
   #
   %x7%e6.%m6 [%5=%4+%2%T3],%0")

;;
;; [rx=ry+i]
;;
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(match_operator:HI
	 6 "plus_or_bound_operator"
	 [(match_operator:HI
	   5 "extend_operator"
	   [(mem:QI (plus:SI
		     (match_operand:SI 2 "bdap_operand" "%r,r,r")
		     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))])
	  (match_operand:HI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[5]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x6%e5.%m5 [%4=%2%S3],%0\";
}")

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 6 "plus_or_bound_operator"
	 [(match_operator:SI
	   5 "extend_operator"
	   [(mem:QI (plus:SI
		     (match_operand:SI 2 "bdap_operand" "%r,r,r")
		     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))])
	  (match_operand:SI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[5]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x6%e5.%m5 [%4=%2%S3],%0\";
}")

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(match_operator:SI
	 6 "plus_or_bound_operator"
	 [(match_operator:SI
	   5 "extend_operator"
	   [(mem:HI (plus:SI
		     (match_operand:SI 2 "bdap_operand" "%r,r,r")
		     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))])
	  (match_operand:SI 1 "register_operand" "0,0,0")]))
   (set (match_operand:SI 4 "register_operand" "=2,r,r")
	(plus:SI (match_dup 2)
		 (match_dup 3)))]
  "(GET_CODE (operands[6]) != UMIN || GET_CODE (operands[5]) == ZERO_EXTEND)
   && side_effect_mode_ok (PLUS, operands, 4, 2, 3, -1, 0)"
  "*
{
  if (which_alternative == 0
      && (GET_CODE (operands[3]) != CONST_INT
	  || INTVAL (operands[3]) > 127
	  || INTVAL (operands[3]) < -128
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'N')
	  || CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'J')))
    return \"#\";
  return \"%x6%e5.%m5 [%4=%2%S3],%0\";
}")

;;
;; Extend versions (zero/sign) of normal add/sub.
;;
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r")
	(match_operator:HI
	 3 "operand_extend_operator"
	 [(match_operand:HI 1 "register_operand" "0,0,0,r")
	  (match_operator:HI
	   4 "extend_operator"
	   [(match_operand:QI 2 "nonimmediate_operand" "r,Q>,m,!To")])]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && (operands[1] != frame_pointer_rtx || GET_CODE (operands[3]) != PLUS)"
  "@
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")
   (set_attr "cc" "clobber")])

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operator:SI
	 3 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0,0,r")
	  (match_operator:SI
	   4 "extend_operator"
	   [(match_operand:QI 2 "nonimmediate_operand" "r,Q>,m,!To")])]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && (operands[1] != frame_pointer_rtx || GET_CODE (operands[3]) != PLUS)"
  "@
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")])

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operator:SI
	 3 "operand_extend_operator"
	 [(match_operand:SI 1 "register_operand" "0,0,0,r")
	  (match_operator:SI
	   4 "extend_operator"
	   [(match_operand:HI 2 "nonimmediate_operand" "r,Q>,m,!To")])]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && (operands[1] != frame_pointer_rtx || GET_CODE (operands[3]) != PLUS)"
  "@
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%0
   %x3%e4.%m4 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")])

;;
;; As with side-effect pattern, may have to have swapped operands for add.
;;
;; ??? Maybe remove this; it *should* be redundant to gcc.
;; QI-mode -> HI-mode
(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r")
	(match_operator:HI
	 4 "plus_or_bound_operator"
	 [(match_operator:HI
	   3 "extend_operator"
	   [(match_operand:QI 2 "nonimmediate_operand" "r,Q>,m,!To")])
	  (match_operand:HI 1 "register_operand" "0,0,0,r")]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && operands[1] != frame_pointer_rtx"
  "@
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")
   (set_attr "cc" "clobber")])

;; QI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operator:SI
	 4 "plus_or_bound_operator"
	 [(match_operator:SI
	   3 "extend_operator"
	   [(match_operand:QI 2 "nonimmediate_operand" "r,Q>,m,!To")])
	  (match_operand:SI 1 "register_operand" "0,0,0,r")]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && operands[1] != frame_pointer_rtx"
  "@
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")])

;; HI-mode -> SI-mode
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operator:SI
	 4 "plus_or_bound_operator"
	 [(match_operator:SI
	   3 "extend_operator"
	   [(match_operand:HI 2 "nonimmediate_operand" "r,Q>,m,!To")])
	  (match_operand:SI 1 "register_operand" "0,0,0,r")]))]
  "(GET_CODE (operands[3]) != UMIN || GET_CODE (operands[4]) == ZERO_EXTEND)
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && operands[1] != frame_pointer_rtx"
  "@
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%0
   %x4%e3.%m3 %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,no")])

;; This is the special case when we use the instruction above in
;; "casesi".  Do not change it to using the generic pattern and
;; "REG 15" as pc, I did that and it led to madness and maintenance
;; problems: Instead of (as imagined) recognizing and removing or
;; replacing this pattern with something simpler, other variant patterns
;; were recognized or combined, including some prefix variants where the
;; value in pc is not that of the next instruction (which means this
;; instruction actually *is* special and *should* be marked as such).
;;  That's what happens when you lie to gcc.  When switching from the
;; (complete) "generic pattern match" approach to this simpler
;; approach, there were insignificant differences in gcc, ipps and
;; product code (somehow due to scratching reload behind the ear or
;; something - testcase "gcc" looked .01% slower and 4 bytes bigger(?);
;; product code became .001% smaller but "looked better".  The testcase
;; "ipps" was just different at register allocation).
(define_insn ""
  [(set (pc)
	(plus:SI
	 (sign_extend:SI
	  (mem:HI (plus:SI
		   (mult:SI (match_operand:SI 0 "register_operand" "r")
			    (const_int 2))
		   (pc))))
	 (pc)))
   (use (label_ref (match_operand:SI 1 "" "")))
   (use (label_ref (match_operand:SI 2 "" "")))
   (use (match_operand 3 "" ""))]

  "operands[0] != frame_pointer_rtx"

  "adds.w [pc+%0.w],pc"
  [(set_attr "cc" "clobber")])

;;
;; Multiply instructions.
;;

;;
;; Sometimes powers of 2 (which are normally canonicalized to a
;; left-shift) appear here, as a result of address reloading.
;; As a special, for values 3 and 5, we can match with an addi, so add those.
;;
;; ??? This may be obsoleted now.
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(mult:SI
	 (match_operand:SI 1 "register_operand" "%0")
	 (match_operand:SI 2 "const_int_operand" "n")))]
  "operands[0] != frame_pointer_rtx
   && operands[1] != frame_pointer_rtx
   && GET_CODE (operands[2]) == CONST_INT
   && (INTVAL (operands[2]) == 2
       || INTVAL (operands[2]) == 4 || INTVAL (operands[2]) == 3
       || INTVAL (operands[2]) == 5)"
  "*
{
  if (INTVAL (operands[2]) == 2)
    return \"lslq 1,%0\";
  else if (INTVAL (operands[2]) == 4)
    return \"lslq 2,%0\";
  else if (INTVAL (operands[2]) == 3)
    return \"addi %0.w,%0\";
  else if (INTVAL (operands[2]) == 5)
    return \"addi %0.d,%0\";
  return \"BAD: adr_mulsi: %0=%1*%2\";
}"
[(set_attr "slottable" "yes")
 (set_attr "cc" "clobber")])

;;
;; Define the addi insn as it normally is used.
;;
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI
	 (mult:SI (match_operand:SI 2 "register_operand" "r")
		  (match_operand:SI 3 "const_int_operand" "n"))
	 (match_operand:SI 1 "register_operand" "0")))]
  "operands[0] != frame_pointer_rtx
   && operands[1] != frame_pointer_rtx
   && GET_CODE (operands[3]) == CONST_INT
   && (INTVAL (operands[3]) == 1
       || INTVAL (operands[3]) == 2 || INTVAL (operands[3]) == 4)"
  "addi %2%T3,%0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

;;
;; The mstep instruction - probably not useful per se, but we will
;; expand to it.
;;
(define_insn "mstep_shift"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(if_then_else
	 (lt (cc0) (const_int 0))
	 (plus:SI (ashift:SI (match_operand:SI 1 "register_operand" "0")
			     (const_int 1))
		  (match_operand:SI 2 "register_operand" "r"))
	 (ashift:SI (match_operand:SI 3 "register_operand" "0")
		    (const_int 1))))]
  ""
  "mstep %2,%0"
  [(set_attr "slottable" "yes")])

;;
;; When illegitimate addresses are legitimized, sometimes gcc forgets
;; to canonicalize the multiplies.
;;
;; ??? Maybe fix gcc or remove this pattern for gcc > 2.7.2.
(define_insn "mstep_mul"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(if_then_else
	 (lt (cc0) (const_int 0))
	 (plus:SI (mult:SI (match_operand:SI 1 "register_operand" "0")
			   (const_int 2))
		  (match_operand:SI 2 "register_operand" "r"))
	 (mult:SI (match_operand:SI 3 "register_operand" "0")
		  (const_int 2))))]
  "operands[0] != frame_pointer_rtx
   && operands[1] != frame_pointer_rtx
   && operands[2] != frame_pointer_rtx
   && operands[3] != frame_pointer_rtx"
  "mstep %2,%0"
  [(set_attr "slottable" "yes")])

;;
;; "Real" multiply - expand into msteps.  Only suitable for extremely
;; time-critical code and the fixedpoint library.
;;
(define_expand "umulhisi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:HI 2 "nonimmediate_operand" "rm")))
   (set (match_dup 4)
	(subreg:SI (match_operand:HI 1 "general_operand" "g") 0))
   (set (match_dup 5) (ashift:SI (match_dup 4)
				 (const_int 16)))
   (set (match_dup 6) (match_dup 5))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (match_operand:SI 0 "nonimmediate_operand" "=rm")
	(match_dup 6))]
  "TARGET_EXPAND_MUL"
  "
{
  int i;

  for (i = 3; i <= 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

;; ??? Unused?
(define_expand "umulqihi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:QI 2 "nonimmediate_operand" "rm")))
   (set (match_dup 4)
	(subreg:SI (match_operand:QI 1 "general_operand" "g") 0))
   (set (match_dup 5) (ashift:SI (match_dup 4)
				 (const_int 24)))
   (set (match_dup 6) (match_dup 5))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(subreg:HI (match_dup 6) 0))]
  "TARGET_EXPAND_MUL"
  "
{
  int i;

  for (i = 3; i <= 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

;;
;; ??? Can gcc expand from "umulhisi3", as stated in optabs.c?
;;
(define_expand "mulsi3"
  [(set (match_operand:SI 0 "general_operand" "=r")
	(mult:SI (match_operand:SI 1 "general_operand" "%0")
		 (match_operand:SI 2 "general_operand" "r")))]
  "TARGET_EXPAND_MUL"
  "
{
  rtx ops[8];
  int i;

  for (i = 0; i < 3; i++)
    ops[i] = gen_reg_rtx (SImode);

  for (i = 3; i < 6; i++)
    ops[i] = gen_reg_rtx (HImode);

  for (i = 6; i < 8; i++)
    ops[i] = gen_reg_rtx (SImode);

  emit_insn (gen_umulhisi3 (ops[0],
			    gen_rtx (SUBREG, HImode, operands[1], 0),
			    gen_rtx (SUBREG, HImode, operands[2], 0)));

  emit_insn (gen_lshrsi3 (ops[1],
			  operands[1],
			  gen_rtx (CONST_INT, VOIDmode, 16)));

  emit_insn (gen_lshrsi3 (ops[2],
			  operands[2],
			  gen_rtx (CONST_INT, VOIDmode, 16)));

  emit_insn (gen_mulhi3 (ops[3],
			 gen_rtx (SUBREG, HImode, ops[1], 0),
			 gen_rtx (SUBREG, HImode, operands[2], 0)));

  emit_insn (gen_mulhi3 (ops[4],
			 gen_rtx (SUBREG, HImode, ops[2], 0),
			 gen_rtx (SUBREG, HImode, operands[1], 0)));

  emit_insn (gen_addhi3 (ops[5], ops[3], ops[4]));

  emit_insn (gen_ashlsi3 (ops[6],
			  gen_rtx (SUBREG, SImode, ops[5], 0),
			  gen_rtx (CONST_INT, VOIDmode, 16)));

  emit_insn (gen_addsi3 (ops[7], ops[0], ops[6]));

  emit_insn (gen_movsi (operands[0], ops[7]));

  DONE;
}")

;;
;; ??? Unused?
;;
(define_expand "mulhi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:HI 2 "nonimmediate_operand" "rm")))
   (set (match_dup 4)
	(subreg:SI (match_operand:HI 1 "general_operand" "g") 0))
   (set (match_dup 5) (ashift:SI (match_dup 4)
				 (const_int 16)))
   (set (match_dup 6) (match_dup 5))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (match_operand:HI 0 "nonimmediate_operand" "=rm")
	(subreg:HI (match_dup 6) 0))]
  "TARGET_EXPAND_MUL"
  "
{
  int i;

  for (i = 3; i <= 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

(define_expand "mulqi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:QI 2 "nonimmediate_operand" "rm")))
   (set (match_dup 4)
	(subreg:SI (match_operand:QI 1 "general_operand" "g") 0))
   (set (match_dup 5) (ashift:SI (match_dup 4)
				 (const_int 24)))
   (set (match_dup 6) (match_dup 5))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (cc0) (match_dup 6))
   (set (match_dup 6) (if_then_else
		       (lt (cc0) (const_int 0))
		       (plus:SI (ashift:SI (match_dup 6)
					   (const_int 1))
				(match_dup 3))
		       (ashift:SI (match_dup 6)
				  (const_int 1))))
   (set (match_operand:QI 0 "nonimmediate_operand" "=rm")
	(subreg:QI (match_dup 6) 0))]
  "TARGET_EXPAND_MUL"
  "
{
  int i;

  for (i = 3; i <= 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

;;
;; Divide and modulus instructions.  CRIS only has a step instruction.
;;
;; First, define the dstep insn.
(define_insn "dstep_shift"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(if_then_else
	 (geu (ashift:SI (match_operand:SI 1 "register_operand" "0")
			 (const_int 1))
	      (match_operand:SI 2 "register_operand" "r"))
	 (minus:SI (ashift:SI (match_operand:SI 3 "register_operand" "0")
			(const_int 1))
		   (match_operand:SI 4 "register_operand" "2"))
	 (ashift:SI (match_operand:SI 5 "register_operand" "0")
			(const_int 1))))]
  ""
  "dstep %2,%0"
  [(set_attr "slottable" "yes")])

;;
;; And a variant with mult instead of ashift.
;;
;; ??? This should be investigated.  Which one matches through combination?
(define_insn "dstep_mul"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(if_then_else
	 (geu (mult:SI (match_operand:SI 1 "register_operand" "0")
		       (const_int 2))
	      (match_operand:SI 2 "register_operand" "r"))
	 (minus:SI (mult:SI (match_operand:SI 3 "register_operand" "0")
			    (const_int 2))
		   (match_operand:SI 4 "register_operand" "2"))
	 (mult:SI (match_operand:SI 5 "register_operand" "0")
		  (const_int 2))))]
  "operands[0] != frame_pointer_rtx
   && operands[1] != frame_pointer_rtx
   && operands[2] != frame_pointer_rtx
   && operands[3] != frame_pointer_rtx"
  "dstep %2,%0"
  [(set_attr "slottable" "yes")])

;;
;; Logical operators.
;;
;; Bitwise "and".
;;

;; There is no use in defining "anddi3", because gcc can expand this by
;; itself, and make (reasonably) good code without interference.

;;
;; If the first operand is memory or a register and is the same as the
;; second operand, and the third operand is -256 or -65536, we can clear
;; instead.  Or if the first operand is a register, and the third
;; operand is 255 or 65535, we can zero_extend. 
;; Gcc isnt smart enough to recognize these cases (yet), and they seem
;; to be common enough to be worthwhile.
;;
;; ??? This should be made obsolete, but I doubt it is (yet).
;;
(define_expand "andsi3"
  [(set (match_operand:SI 0 "nonimmediate_operand"	   "")
	(and:SI (match_operand:SI 1 "nonimmediate_operand" "")
		(match_operand:SI 2 "general_operand"	 "")))]
  ""
  "
{
  if (! (GET_CODE (operands[2]) == CONST_INT
	 && (((INTVAL (operands[2]) == -256
	       || INTVAL (operands[2]) == -65536)
	      && rtx_equal_p (operands[1], operands[0]))
	     || ((INTVAL (operands[2]) == 255
		  || INTVAL (operands[2]) == 65535)
		 && REG_P (operands[0])))))
    {
      /* Make intermediate steps if operand0 is not a register or
	 operand1 is not a register, and hope that the reload pass will
	 make something useful out of it.  Note that the operands are
	 *not* canonicalized.  For the 
	 moment, I chicken out on this, because:
	 - All or most ports do not describe 'and' with canonicalized
	   operands.
	 - I seem to remember magic in reload, checking that operand1 has
	   constraint '%0', in which case operand0 and operand1 must have
	   similar predicates.
	 - Kva, kva...
	 */
      rtx reg0 = REG_P (operands[0]) ? operands[0] : gen_reg_rtx (SImode);
      rtx reg1 = operands[1];

      if (! REG_P (reg1))
	{
	  emit_move_insn (reg0, reg1);
	  reg1 = reg0;
	}

      emit_insn (gen_rtx (SET, SImode, reg0,
			  gen_rtx (AND, SImode, reg1, operands[2])));

      /* Make sure we get the right *final* destination. */
      if (! REG_P (operands[0]))
	emit_move_insn (operands[0], reg0);

      DONE;
    }
}")

(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r")
	(and:SI (match_operand:SI 1 "nonimmediate_operand" "%r,Q>,m")
		(match_operand:SI 2 "const_int_operand" "n,n,n")))]
  "INTVAL (operands[2]) == 255 || INTVAL (operands[2]) == 65535"
  "movu.%z2 %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn ""
  [(set (match_operand:SI 0 "nonimmediate_operand" "=r,r,Q>,Q>,m,m")
	(and:SI (match_operand:SI 1 "nonimmediate_operand" "%0,0,0,0,0,0")
		(match_operand:SI 2 "const_int_operand" "P,n,P,n,P,n")))]
  "INTVAL (operands[2]) == -65536 || INTVAL (operands[2]) == -256"
  "@
   cLear.b %0
   cLear.w %0
   cLear.b %0
   cLear.w %0
   cLear.b %0
   cLear.w %0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "none")])

;;
;; This pattern has to take care of everything that was not matched in
;; the insns above.
;;
;; Sidenote: the tightening from "nonimmediate_operand" to
;; "register_operand" for operand 1 actually increased the register
;; pressure (worse code).  That will hopefully change with an
;; improved reload pass.
(define_insn ""
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r,r")
	(and:SI (match_operand:SI 1 "register_operand" "%0,0,0,0,r")
		(match_operand:SI 2 "general_operand" "I,r,Q>,g,!To")))]
  ""
  "@
   andq %2,%0
   and.d %2,%0
   and.d %2,%0
   and.d %2,%0
   and.d %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,no,no")])

;;
;; For both QI and HI we may use the quick patterns.  This results in
;; useless condition codes, but that is used rarely enough for it to
;; normally be a win (could check ahead for use of cc0, but seems to be
;; more pain than win).
;;

;; See note for andsi3
(define_expand "andhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
	(and:HI (match_operand:HI 1 "nonimmediate_operand" "")
		(match_operand:HI 2 "general_operand"  "")))]
  ""
  "
{
  if (! (GET_CODE (operands[2]) == CONST_INT
	 && (((INTVAL (operands[2]) == -256
	       || INTVAL (operands[2]) == 65280)
	      && rtx_equal_p (operands[1], operands[0]))
	     || (INTVAL (operands[2]) == 255
		 && REG_P (operands[0])))))
    {
      /* See comment for andsi3. */
      rtx reg0 = REG_P (operands[0]) ? operands[0] : gen_reg_rtx (HImode);
      rtx reg1 = operands[1];

      if (! REG_P (reg1))
	{
	  emit_move_insn (reg0, reg1);
	  reg1 = reg0;
	}

      emit_insn (gen_rtx (SET, HImode, reg0,
			  gen_rtx (AND, HImode, reg1, operands[2])));

      /* Make sure we get the right destination. */
      if (! REG_P (operands[0]))
	emit_move_insn (operands[0], reg0);

      DONE;
    }
}")

(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r")
	(and:HI (match_operand:HI 1 "nonimmediate_operand" "r,Q>,m")
		(const_int 255)))]
  ""
  "mOvu.b %1,%0"
  [(set_attr "slottable" "yes,yes,no")])

(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r,Q>,m")
	(and:HI (match_operand:HI 1 "nonimmediate_operand" "0,0,0")
		(const_int -256)))]
  ""
  "cLear.b %0"
  [(set_attr "slottable" "yes,yes,no")
   (set_attr "cc" "none")])

;; ??? One of these two patterns must be redundant.
(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r,Q>,m")
	(and:HI (match_operand:HI 1 "nonimmediate_operand" "0,0,0")
		(const_int 65280)))]
  ""
  "cLear.b %0"
  [(set_attr "slottable" "yes,yes,no")
   (set_attr "cc" "none")])

(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r,r,r,r")
	(and:HI (match_operand:HI 1 "register_operand" "%0,0,0,0,0,0,r")
		(match_operand:HI 2 "general_operand" "I,r,Q>,L,O,g,!To")))]
;; Sidenote: the tightening from "general_operand" to
;; "register_operand" for operand 1 actually increased the register
;; pressure (worse code).  That will hopefully change with an
;; improved reload pass.
  ""
  "@
   andq %2,%0
   and.w %2,%0
   and.w %2,%0
   and.w %2,%0
   anDq %b2,%0
   and.w %2,%0
   and.w %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,no,yes,no,no")
   (set_attr "cc" "clobber,normal,normal,normal,clobber,normal,normal")])

(define_insn ""
  [(set (strict_low_part
	 (match_operand:HI 0 "register_operand" "=r,r,r,r,r,r"))
	(and:HI (match_operand:HI 1 "register_operand" "%0,0,0,0,0,r")
		(match_operand:HI 2 "general_operand" "r,Q>,L,O,g,!To")))]
  ""
  "@
   and.w %2,%0
   and.w %2,%0
   and.w %2,%0
   anDq %b2,%0
   and.w %2,%0
   and.w %2,%1,%0"
  [(set_attr "slottable" "yes,yes,no,yes,no,no")
   (set_attr "cc" "normal,normal,normal,clobber,normal,normal")])

(define_insn "andqi3"
  [(set (match_operand:QI 0 "register_operand" "=r,r,r,r,r,r")
	(and:QI (match_operand:QI 1 "register_operand" "%0,0,0,0,0,r")
		(match_operand:QI 2 "general_operand" "I,r,Q>,O,g,!To")))]
  ""
  "@
   andq %2,%0
   and.b %2,%0
   and.b %2,%0
   andQ %b2,%0
   and.b %2,%0
   and.b %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "clobber,normal,normal,clobber,normal,normal")])

(define_insn ""
  [(set (strict_low_part
	 (match_operand:QI 0 "register_operand" "=r,r,r,r,r"))
	(and:QI (match_operand:QI 1 "register_operand" "%0,0,0,0,r")
		(match_operand:QI 2 "general_operand" "r,Q>,O,g,!To")))]
  ""
  "@
   and.b %2,%0
   and.b %2,%0
   andQ %b2,%0
   and.b %2,%0
   and.b %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,no,no")
   (set_attr "cc" "normal,normal,clobber,normal,normal")])

;;
;; Bitwise or.
;;

;;
;; Same comment as anddi3 applies here - no need for such a pattern.
;;

(define_insn "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r,r,r")
	(ior:SI (match_operand:SI 1 "register_operand" "%0,0,0,0,0,r")
		(match_operand:SI 2 "general_operand" "I,r,Q>,n,g,!To")))]
  ""
  "@
   orq %2,%0
   or.d %2,%0
   or.d %2,%0
   oR.%s2 %2,%0
   or.d %2,%0
   or.d %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,no,no,no")
   (set_attr "cc" "normal,normal,normal,clobber,normal,normal")])

(define_insn "iorhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r,r,r,r,r,r")
	(ior:HI (match_operand:HI 1 "register_operand" "%0,0,0,0,0,0,r")
		(match_operand:HI 2 "general_operand" "I,r,Q>,L,O,g,!To")))]
  ""
  "@
   orq %2,%0
   or.w %2,%0
   or.w %2,%0
   or.w %2,%0
   oRq %b2,%0
   or.w %2,%0
   or.w %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,no,yes,no,no")
   (set_attr "cc" "clobber,normal,normal,normal,clobber,normal,normal")])

(define_insn "iorqi3"
  [(set (match_operand:QI 0 "register_operand" "=r,r,r,r,r,r")
	(ior:QI (match_operand:QI 1 "register_operand" "%0,0,0,0,0,r")
		(match_operand:QI 2 "general_operand" "I,r,Q>,O,g,!To")))]
  ""
  "@
   orq %2,%0
   or.b %2,%0
   or.b %2,%0
   orQ %b2,%0
   or.b %2,%0
   or.b %2,%1,%0"
  [(set_attr "slottable" "yes,yes,yes,yes,no,no")
   (set_attr "cc" "clobber,normal,normal,clobber,normal,normal")])

;;
;; xor
;;

;; See comment about "anddi3" for xordi3 - no need for such a pattern.

(define_insn "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(xor:SI (match_operand:SI 1 "register_operand" "%0")
		(match_operand:SI 2 "register_operand" "r")))]
  ""
  "xor %2,%0"
  [(set_attr "slottable" "yes")])

(define_insn "xorhi3"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(xor:HI (match_operand:HI 1 "register_operand" "%0")
		(match_operand:HI 2 "register_operand" "r")))]
  ""
  "xor %2,%0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "clobber")])

(define_insn "xorqi3"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(xor:QI (match_operand:QI 1 "register_operand" "%0")
		(match_operand:QI 2 "register_operand" "r")))]
  ""
  "xor %2,%0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "clobber")])

;;
;; Negation insns.
;;

;;
;; Just here as a (slightly usable) example.
;;
(define_expand "negsf2"
  [(set (match_dup 2)
        (match_dup 3))
   (parallel [(set (match_operand:SF 0 "register_operand" "=r")
                   (neg:SF (match_operand:SF 1 
                            "register_operand" "0")))
              (use (match_dup 2))])]
  ""
  "
{
  operands[2] = gen_reg_rtx (SImode);
  operands[3] = GEN_INT (0x80000000);
}")

(define_insn ""
  [(set (match_operand:SF 0 "register_operand" "=r")
	(neg:SF (match_operand:SF 1 "register_operand" "0")))
   (use (match_operand:SI 2 "register_operand" "r"))]
  ""
  "xor %2,%0"
  [(set_attr "slottable" "yes")])

;;
;; No "negdi2" although we *could* make one up that may be faster than
;; the one lin libgcc.
;;

(define_insn "negsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(neg:SI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "neg.d %1,%0"
  [(set_attr "slottable" "yes")])

(define_insn "neghi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(neg:HI (match_operand:HI 1 "register_operand" "r")))]
  ""
  "neg.w %1,%0"
  [(set_attr "slottable" "yes")])

(define_insn "negqi2"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(neg:QI (match_operand:QI 1 "register_operand" "r")))]
  ""
  "neg.b %1,%0"
  [(set_attr "slottable" "yes")])

;;
;; One-complements
;;
;; See comment on anddi3 - no need for a DImode pattern.
;;
(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (match_operand:SI 1 "register_operand" "0")))]
  ""
  "not %0"
  [(set_attr "slottable" "yes")])

(define_insn "one_cmplhi2"
  [(set (match_operand:HI 0 "register_operand" "=r")
	(not:HI (match_operand:HI 1 "register_operand" "0")))]
  ""
  "not %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "clobber")])

(define_insn "one_cmplqi2"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(not:QI (match_operand:QI 1 "register_operand" "0")))]
  ""
  "not %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "clobber")])

;;
;; Arithmetic shift right
;;
(define_insn "ashrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashiftrt:SI (match_operand:SI 1 "register_operand" "0")
		     (match_operand:SI 2 "nonmemory_operand" "nr")))]
  ""
  "*
{
  if (REG_S_P (operands[2]))
    return \"asr.d %2,%0\";

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) > 31)
    return \"moveq -1,%0\";
  else
    return \"asrq %2,%0\";
}"
  [(set_attr "slottable" "yes")])

;;
;; Since gcc gets lost, and forgets to zero-extend the source (or mask
;; the destination) when it changes shifts of lower modes into SImode,
;; it is better to make these expands and anonymous patterns instead of
;; the more correct define_insns.  This occurs when gcc thinks that is
;; is better to widen to SImode and use immediate shift count.
;; ??? Is this legacy or still true for gcc >= 2.7.2
(define_expand "ashrhi3"
  [(set (match_dup 3)
	(sign_extend:SI (match_operand:HI 1 "nonimmediate_operand" "rm")))
   (set (match_dup 4)
	(zero_extend:SI (match_operand:HI 2 "nonimmediate_operand" "rm")))
   (set (match_dup 5) (ashiftrt:SI (match_dup 3) (match_dup 4)))
   (set (match_operand:HI 0 "general_operand" "=g")
	(subreg:HI (match_dup 5) 0))]
  ""
  "
{
  int i;

  for (i = 3; i < 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r")
	(ashiftrt:HI (match_operand:HI 1 "register_operand" "0")
		     (match_operand:HI 2 "register_operand" "r")))]
  ""
  "asr.w %2,%0"
  [(set_attr "slottable" "yes")])

(define_insn ""
  [(set (strict_low_part (match_operand:HI 0 "register_operand" "+r"))
	(ashiftrt:HI (match_dup 0)
		     (match_operand:HI 1 "register_operand" "r")))]
  ""
  "asr.w %1,%0"
  [(set_attr "slottable" "yes")])

;; Same comment goes as for "ashrhi3".
(define_expand "ashrqi3"
  [(set (match_dup 3)
	(sign_extend:SI (match_operand:QI 1 "nonimmediate_operand" "g")))
   (set (match_dup 4)
	(zero_extend:SI (match_operand:QI 2 "nonimmediate_operand" "g")))
   (set (match_dup 5) (ashiftrt:SI (match_dup 3) (match_dup 4)))
   (set (match_operand:QI 0 "general_operand" "=g")
	(subreg:QI (match_dup 5) 0))]
  ""
  "
{
  int i;

  for (i = 3; i < 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r")
	(ashiftrt:QI (match_operand:QI 1 "register_operand" "0")
		     (match_operand:QI 2 "register_operand" "r")))]
  ""
  "asr.b %2,%0"
  [(set_attr "slottable" "yes")])

(define_insn ""
  [(set (strict_low_part (match_operand:QI 0 "register_operand" "+r"))
	(ashiftrt:QI (match_dup 0)
		     (match_operand:QI 1 "register_operand" "r")))]
  ""
  "asr.b %1,%0"
  [(set_attr "slottable" "yes")])

;;
;; Logical shift right
;;
(define_insn "lshrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lshiftrt:SI (match_operand:SI 1 "register_operand" "0")
		     (match_operand:SI 2 "nonmemory_operand" "nr")))]
  ""
  "*
{
  if (REG_S_P (operands[2]))
    return \"lsr.d %2,%0\";

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) > 31)
    return \"moveq 0,%0\";
  else
    return \"lsrq %2,%0\";
}"
  [(set_attr "slottable" "yes")])

;; Same comments as for ashrhi3
(define_expand "lshrhi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:HI 1 "nonimmediate_operand" "g")))
   (set (match_dup 4)
	(zero_extend:SI (match_operand:HI 2 "nonimmediate_operand" "g")))
   (set (match_dup 5) (lshiftrt:SI (match_dup 3) (match_dup 4)))
   (set (match_operand:HI 0 "general_operand" "=g")
	(subreg:HI (match_dup 5) 0))]
  ""
  "
{
  int i;

  for (i = 3; i < 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

(define_insn ""
  [(set (match_operand:HI 0 "register_operand" "=r")
	(lshiftrt:HI (match_operand:HI 1 "register_operand" "0")
		     (match_operand:HI 2 "register_operand" "r")))]
  ""
  "lsr.w %2,%0"
  [(set_attr "slottable" "yes")])

(define_insn ""
  [(set (strict_low_part (match_operand:HI 0 "register_operand" "+r"))
	(lshiftrt:HI (match_dup 0)
		     (match_operand:HI 1 "register_operand" "r")))]
  ""
  "lsr.w %1,%0"
  [(set_attr "slottable" "yes")])

;; Same comments as for ashrhi3
(define_expand "lshrqi3"
  [(set (match_dup 3)
	(zero_extend:SI (match_operand:QI 1 "nonimmediate_operand" "g")))
   (set (match_dup 4)
	(zero_extend:SI (match_operand:QI 2 "nonimmediate_operand" "g")))
   (set (match_dup 5) (lshiftrt:SI (match_dup 3) (match_dup 4)))
   (set (match_operand:QI 0 "general_operand" "=g")
	(subreg:QI (match_dup 5) 0))]
  ""
  "
{
  int i;

  for (i = 3; i < 6; i++)
    operands[i] = gen_reg_rtx (SImode);
}")

(define_insn ""
  [(set (match_operand:QI 0 "register_operand" "=r")
	(lshiftrt:QI (match_operand:QI 1 "register_operand" "0")
		     (match_operand:QI 2 "register_operand" "r")))]
  ""
  "lsr.b %2,%0"
  [(set_attr "slottable" "yes")])

(define_insn ""
  [(set (strict_low_part (match_operand:QI 0 "register_operand" "+r"))
	(lshiftrt:QI (match_dup 0)
		     (match_operand:QI 1 "register_operand" "r")))]
  ""
  "lsr.b %1,%0"
  [(set_attr "slottable" "yes")])

;;
;; Arithmetic/logical shift left
;;
(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ashift:SI (match_operand:SI 1 "register_operand" "0")
		   (match_operand:SI 2 "nonmemory_operand" "nr")))]
  ""
  "*
{
  if (REG_S_P (operands[2]))
    return \"lsl.d %2,%0\";

  if (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) > 31)
    return \"moveq 0,%0\";
  else
    return \"lslq %2,%0\";
}"
  [(set_attr "slottable" "yes")])

;;
;; For narrower modes than SI, we may use lslq although it makes cc
;; unusable, but we do not have to reload the shift-count into a register.
;;
(define_insn "ashlhi3"
  [(set (match_operand:HI 0 "register_operand" "=r,r")
	(ashift:HI (match_operand:HI 1 "register_operand" "0,0")
		   (match_operand:HI 2 "nonmemory_operand" "r,n")))]

  ""
  "*
{
  return
    (GET_CODE (operands[2]) == CONST_INT && INTVAL (operands[2]) > 15)
    ? \"moveq 0,%0\"
    : (CONSTANT_P (operands[2])
       ? \"lslq %2,%0\" : \"lsl.w %2,%0\");
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "normal,clobber")])

(define_insn ""
  [(set (strict_low_part (match_operand:HI 0 "register_operand" "+r"))
	(ashift:HI (match_dup 0)
		   (match_operand:HI 1 "register_operand" "r")))]
  ""
  "lsl.w %1,%0"
  [(set_attr "slottable" "yes")])

(define_insn "ashlqi3"
  [(set (match_operand:QI 0 "register_operand" "=r,r")
	(ashift:QI (match_operand:QI 1 "register_operand" "0,0")
		   (match_operand:QI 2 "nonmemory_operand" "r,n")))]
  ""
  "*
{
  return
    (GET_CODE (operands[2]) == CONST_INT
     && INTVAL (operands[2]) > 7)
    ? \"moveq 0,%0\"
    : (CONSTANT_P (operands[2])
       ? \"lslq %2,%0\" : \"lsl.b %2,%0\");
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "normal,clobber")])

(define_insn ""
  [(set (strict_low_part (match_operand:QI 0 "register_operand" "+r"))
	(ashift:QI (match_dup 0)
		   (match_operand:QI 1 "register_operand" "r")))]
  ""
  "lsl.b %1,%0"
  [(set_attr "slottable" "yes")])

;;
;; Various strange insns that gcc likes.
;;

;;
;; Fortunately, it is simple to construct an abssf (although it may not
;; be very much used in practice).
;;
(define_insn "abssf2"
  [(set (match_operand:SF 0 "register_operand" "=r")
	(abs:SF (match_operand:SF 1 "register_operand" "0")))]
  ""
  "lslq 1,%0\;lsrq 1,%0")

(define_insn "abssi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(abs:SI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "abs %1,%0"
  [(set_attr "slottable" "yes")])

;; ??? Gcc should be able to do these expansions itself.

(define_expand "abshi2"
  [(set (match_dup 2)
	(sign_extend:SI (match_operand:HI 1 "general_operand" "g")))
   (set (match_dup 3) (abs:SI (match_dup 2)))
   (set (match_operand:HI 0 "register_operand" "=r")
	(subreg:HI (match_dup 3) 0))]
  ""
  "operands[2] = gen_reg_rtx (SImode); operands[3] = gen_reg_rtx (SImode);")

(define_expand "absqi2"
  [(set (match_dup 2)
	(sign_extend:SI (match_operand:QI 1 "general_operand" "g")))
   (set (match_dup 3) (abs:SI (match_dup 2)))
   (set (match_operand:QI 0 "register_operand" "=r")
	(subreg:QI (match_dup 3) 0))]
  ""
  "operands[2] = gen_reg_rtx (SImode); operands[3] = gen_reg_rtx (SImode);")

;;
;; Bound-insn.	Arithmetically defined to be the same as an unsigned
;; minimum, which is an operation supported by gcc.  Used in casesi, but
;; seems to be used now and then in real code too.
;;
(define_insn "uminsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(umin:SI  (match_operand:SI 1 "register_operand" "%0,0,0,r")
		  (match_operand:SI 2 "nonmemory_operand" "r,Q>,g,!To")))]
  ""
  "*
{
  if (GET_CODE (operands[2]) == CONST_INT)
    {
      if (INTVAL (operands[2]) < 256)
	return \"bound.b %2,%0\";

      if (INTVAL (operands[2]) < 65536)
	return \"bound.w %2,%0\";
    }
  else if (which_alternative == 3)
    return \"bound.d %2,%1,%0\";

  return \"bound.d %2,%0\";
}"
 [(set_attr "slottable" "yes,yes,no,no")])

;;
;; Jump and branch insns.
;;
(define_insn "jump"
  [(set (pc)
	(label_ref (match_operand 0 "" "")))]
  ""
  "ba %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "indirect_jump"
  [(set (pc) (match_operand:SI 0 "general_operand" "g"))]
  ""
  "jump %0")

;;
;; Return insn - use whenever the epilogue is very simple;
;;  if it is only a single ret or jump [sp+] (no saved registers or
;;  allocated stack space)
;; In this pattern, although named, it is ok to check for the
;; context in the test, not only on compiler switches.
;;
(define_insn "return"
  [(return)]
  "simple_epilogue ()"
  "*
{
  int i;

   /* Just needs to hold a 'movem [sp+],rN'. */
  char rd[16];

  *rd = 0;

  /* Start from the last call-saved register.
      We know that we have a simple epilogue, so we just have to find the
     last register in the movem sequence.  */
  for (i = 8; i >= 0; i--)
    if (regs_ever_live[i])
       break;

  if (i >= 0)
    sprintf (rd, \"movem [sp+],r%d\", i);

  if (regs_ever_live[SRP_REGNUM])
    {
      if (*rd)
	output_asm_insn (rd, operands);
      return \"jump [sp+]\";
    }

  if (*rd)
    {
      output_asm_insn (\"reT\", operands);
      output_asm_insn (rd, operands);
      return \"\";
    }

  return \"ret%#\";
}"
  [(set (attr "slottable")
	(if_then_else
	 (ne (symbol_ref "regs_ever_live[SRP_REGNUM]") (const_int 0))
	 (const_string "no")	     ; If jump then not slottable.
	 (if_then_else
	  (ne (symbol_ref "regs_ever_live[0]") (const_int 0))
	  (const_string "no") ; ret+movem [sp+],rx: slot already filled.
	  (const_string "branch")))) ; If ret then delayed.
   (set_attr "cc" "none")])

;;
;; Conditional branches.
;;
(define_insn "beq"
  [(set (pc)
	(if_then_else (eq (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "beq %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "bne"
  [(set (pc)
	(if_then_else (ne (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "bne %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "bgt"
  [(set (pc)
	(if_then_else (gt (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"bgt %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn "bgtu"
  [(set (pc)
	(if_then_else (gtu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "bhi %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "blt"
  [(set (pc)
	(if_then_else (lt (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"bmi %l0%#\" : \"blt %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn "bltu"
  [(set (pc)
	(if_then_else (ltu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "blo %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "bge"
  [(set (pc)
	(if_then_else (ge (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"bpl %l0%#\" : \"bge %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn "bgeu"
  [(set (pc)
	(if_then_else (geu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "bhs %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn "ble"
  [(set (pc)
	(if_then_else (le (cc0)
			  (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"ble %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn "bleu"
  [(set (pc)
	(if_then_else (leu (cc0)
			   (const_int 0))
		      (label_ref (match_operand 0 "" ""))
		      (pc)))]
  ""
  "bls %l0%#"
  [(set_attr "slottable" "branch")])

;;
;; Reversed anonymous patterns to the ones above.
;;
(define_insn ""
  [(set (pc)
	(if_then_else (eq (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "bne %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (ne (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "beq %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (gt (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"ble %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (gtu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "bls %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (lt (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"bpl %l0%#\" : \"bge %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (ltu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "bhs %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (ge (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"bmi %l0%#\" : \"blt %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (geu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "blo %l0%#"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (le (cc0)
			  (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"bgt %l0%#\";
}"
  [(set_attr "slottable" "branch")])

(define_insn ""
  [(set (pc)
	(if_then_else (leu (cc0)
			   (const_int 0))
		      (pc)
		      (label_ref (match_operand 0 "" ""))))]
  ""
  "bhi %l0%#"
  [(set_attr "slottable" "branch")])

;;
;; Set on condition : Scc
;;
(define_insn "sgeu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(geu (cc0) (const_int 0)))]
  ""
  "shs %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sltu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ltu (cc0) (const_int 0)))]
  ""
  "slo %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "seq"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(eq (cc0) (const_int 0)))]
  ""
  "seq %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sge"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ge (cc0) (const_int 0)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"spl %0\" : \"sge %0\";
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sgt"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gt (cc0) (const_int 0)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"sgt %0\";
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sgtu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(gtu (cc0) (const_int 0)))]
  ""
  "shi %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sle"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(le (cc0) (const_int 0)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  0 : \"sle %0\";
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sleu"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(leu (cc0) (const_int 0)))]
  ""
  "sls %0"
  [(set_attr "slottable" "yes")])

(define_insn "slt"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(lt (cc0) (const_int 0)))]
  ""
  "*
{
  return
    (cc_prev_status.flags & CC_NO_OVERFLOW)
    ?  \"smi %0\" : \"slt %0\";
}"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

(define_insn "sne"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(ne (cc0) (const_int 0)))]
  ""
  "sne %0"
  [(set_attr "slottable" "yes")
   (set_attr "cc" "none")])

;;
;; Call insns
;;
;;
;; We need to make these patterns "expand", since the real operand is
;; hidden in a (mem:QI ) inside operand[0] (call_value: operand[1]),
;; and cannot be checked if it were a "normal" pattern.
;;  I believe someone goofed when it comes to this mem-kind hiding,
;; since it totally stops validation of operands on machines where the
;; call-insn does *not* "automatically" strip the (mem) from its operands.
;;  Note that "call" and "call_value" are *always* called with a
;; mem-operand for operand 0 and 1 respective.	What happen for combined
;; instructions are another issue.
;;
(define_expand "call"
  [(parallel [(call (match_operand:QI 0 "memory_operand" "")
		    (match_operand 1 "general_operand" ""))
	      ;; 16 is the srp (can't use the symbolic name here)
	      (clobber (reg:SI 16))])]
  ""
  "")

;;
;; Accept ANYTHING as operand 1.  Accept operands for operand 0 in
;; order of preference (Q includes r, but r is shorter, faster)
;;
(define_insn ""
  [(call (mem:QI (match_operand:SI 0 "general_operand" "r,Q>,g"))
	 (match_operand 1 "" ""))
   (clobber (reg:SI 16))] ;; 16 is the srp (can't use symbolic name)
  ""
  "jsr %0")

(define_expand "call_value"
  [(parallel [(set (match_operand 0 "" "")
		   (call (match_operand:QI 1 "memory_operand" "")
			 (match_operand 2 "" "")))
	      ;; 16 is the srp (can't use symbolic name)
	      (clobber (reg:SI 16))])]
  ""
  "")

;;
;; Accept ANYTHING as operand 2.  The validity other than "general" of
;; operand 0 will be checked elsewhere.	 Accept operands for operand 1 in
;; order of preference (Q includes r, but r is shorter, faster)
;;
(define_insn ""
  [(set (match_operand 0 "general_operand" "=g,g,g")
	(call (mem:QI (match_operand:SI 1 "general_operand" "r,Q>,g"))
	      (match_operand 2 "" "")))
   (clobber (reg:SI 16))]
  ""
  "Jsr %1"
  [(set_attr "cc" "clobber")])

;;
;; Used in debugging.  No use for the direct pattern; unfilled
;; delayed-branches are taken care of by other means.
;;
(define_insn "nop"
  [(const_int 0)]
  ""
  ""
  [(set_attr "cc" "none")])

;; Note: if you change the "parallel" (or add anything after it) in
;; this expansion, you must change the macro ASM_OUTPUT_CASE_END
;; accordingly, to add the default case at the end of the jump-table.
;;
(define_expand "casesi"
  [(set (match_dup 5) (match_operand:SI 0 "general_operand" ""))
   (set (match_dup 6)
	(minus:SI (match_dup 5)
		  (match_operand:SI 1 "const_int_operand" "n")))
   (set (match_dup 7)
	(umin:SI (match_dup 6)
		 (match_operand:SI 2 "const_int_operand" "n")))
   (parallel
    [(set (pc)
	  (plus:SI (sign_extend:SI (mem:HI
				    (plus:SI (mult:SI (match_dup 7)
						      (const_int 2))
					     (pc))))
		 (pc)))
     (use (label_ref (match_operand 3 "" "")))
     (use (label_ref (match_operand 4 "" "")))
     (use (match_dup 7))])]
  ""
  "
{
  operands[2] = plus_constant (operands[2], 1);
  operands[5] = gen_reg_rtx (SImode);
  operands[6] = gen_reg_rtx (SImode);
  operands[7] = gen_reg_rtx (SImode);
}")

;;
;; Splitting patterns.	Some of them have modes unspecified.  This
;; should always be ok.	 (If you don't agree, then sparc.md has it as
;; well).
;;
;; When register_operand is specified for an operand, we can get a
;; subreg as well (Axis-990331), so don't just assume that REG_P is true
;; and REGNO can plainly be used.  It is best to guard with REG_P if it
;; is not worth it to adjust for the subreg case.
;;
;; op [rx],ry,rz
;; Note the absence of index on rx.  This happens
;; when op:ing together some values to end up in rz.

;;
;; Recognize  bound [rx],ry,rz; where [rx] is zero_extended,
;; and add/sub [rx],ry,rz, with zero or sign extend on [r].
;;  Change into move ry,rz; op [rx],rz
;; Lose if rz=ry or rx=rz
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 4 "operand_extend_operator"
	 [(match_operand 1 "register_operand" "r")
	  (match_operator
	   3 "extend_operator"
	   [(match_operand 2 "memory_operand" "m")])]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) != REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 1))
   (set (match_dup 0)
	(match_op_dup
	 4 [(match_dup 0)
	    (match_op_dup 3 [(match_dup 2)])]))]
  "")

;;
;; As above, but change op [rz],ry,rz into ext [rz],rz; op ry,rz
;; plus or bound only.
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 4 "plus_or_bound_operator"
	 [(match_operand 1 "register_operand" "r")
	  (match_operator
	   3 "extend_operator"
	   [(match_operand 2 "memory_operand" "m")])]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) == REGNO (operands[0])"
  [(set (match_dup 0)
	(match_op_dup 3 [(match_dup 2)]))
   (set (match_dup 0)
	(match_op_dup
	 4 [(match_dup 0)
	    (match_dup 1)]))]
  "")

;;
;; As above, but swapped operands for add and bound
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 4 "plus_or_bound_operator"
	 [(match_operator
	   3 "extend_operator"
	   [(match_operand 2 "memory_operand" "m")])
	  (match_operand 1 "register_operand" "r")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) != REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 1))
   (set (match_dup 0)
	(match_op_dup
	 4 [(match_dup 0)
	    (match_op_dup 3 [(match_dup 2)])]))]
  "")

;;
;; As above, but change op [rz],ry,rz into ext [rz],rz; op ry,rz
;; plus or bound only, of course.
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 4 "plus_or_bound_operator"
	 [(match_operator
	   3 "extend_operator"
	   [(match_operand 2 "memory_operand" "m")])
	  (match_operand 1 "register_operand" "r")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) == REGNO (operands[0])"
  [(set (match_dup 0)
	(match_op_dup 3 [(match_dup 2)]))
   (set (match_dup 0)
	(match_op_dup
	 4 [(match_dup 0)
	    (match_dup 1)]))]
  "")

;;
;;
;; Recognize as above, but the mem operand is not extended.
;;
;; op [rx],ry,rz changed into move ry,rz; op [rx],rz
;; lose if ry=rz or rx=rz
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 3 "orthogonal_operator"
	 [(match_operand 1 "register_operand" "r")
	  (match_operand 2 "memory_operand" "m")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) != REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 1))
   (set (match_dup 0)
	(match_op_dup
	 3 [(match_dup 0)
	    (match_dup 2)]))]
  "")

;;
;; As above, but swapped operands.
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 3 "commutative_orth_op"
	 [(match_operand 2 "memory_operand" "m")
	  (match_operand 1 "register_operand" "r")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) != REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 1))
   (set (match_dup 0)
	(match_op_dup
	 3 [(match_dup 0)
	    (match_dup 2)]))]
  "")

;;
;; As above, but take care of op [rz],ry,rz into move [rz],rz; op ry,rz
;; Only commutative operators.
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 3 "commutative_orth_op"
	 [(match_operand 1 "register_operand" "r")
	  (match_operand 2 "memory_operand" "m")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) == REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 2))
   (set (match_dup 0)
	(match_op_dup
	 3 [(match_dup 0)
	    (match_dup 1)]))]
  "")

;;
;; Again as above, swapped operands.
;;
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 3 "orthogonal_operator"
	 [(match_operand 2 "memory_operand" "m")
	  (match_operand 1 "register_operand" "r")]))]
  "REG_P (operands[0]) && REG_P (operands[1])
   && REGNO (operands[1]) != REGNO (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && REG_P (XEXP (operands[2], 0))
   && REGNO (XEXP (operands[2], 0)) == REGNO (operands[0])"
  [(set (match_dup 0)
	(match_dup 2))
   (set (match_dup 0)
	(match_op_dup
	 3 [(match_dup 0)
	    (match_dup 1)]))]
  "")

;;
;; Splits for all cases where (possibly after reload and local/global
;; register allocation) rx and ry in [rx=ry+i] are equal in side-effect insns.
;;
(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	   (mem (plus:SI
		 (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			  (match_operand:SI 2 "const_int_operand" "n,n"))
		 (match_operand:SI 3 "register_operand" "r,r"))))
     (set (match_operand:SI 4 "register_operand" "=3,r")
	   (plus:SI (mult:SI (match_dup 1)
			     (match_dup 2))
		    (match_dup 3)))])]
  "REG_P (operands[3]) && REG_P (operands[4])
   && REGNO (operands[3]) == REGNO (operands[4])"
  [(set (match_dup 4) (plus:SI (mult:SI (match_dup 1) (match_dup 2))
				(match_dup 3)))
   (set (match_dup 0) (match_dup 5))]
  "operands[5] = gen_rtx (MEM, GET_MODE (operands[0]), operands[3]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	   (mem (plus:SI (match_operand:SI 1 "bdap_operand" "%r,r,r")
			 (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri"))))
     (set (match_operand:SI 3 "register_operand" "=1,r,r")
	   (plus:SI (match_dup 1)
		    (match_dup 2)))])]
  "(rtx_equal_p (operands[3], operands[1])
    || rtx_equal_p (operands[3], operands[2]))"
  [(set (match_dup 3) (plus:SI (match_dup 1) (match_dup 2)))
   (set (match_dup 0) (match_dup 4))]
  "operands[4] = gen_rtx (MEM, GET_MODE (operands[0]), operands[3]);")

(define_split
  [(parallel
    [(set (mem (plus:SI
		 (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			  (match_operand:SI 1 "const_int_operand" "n,n"))
		 (match_operand:SI 2 "register_operand" "r,r")))
	   (match_operand 3 "register_operand" "r,r"))
     (set (match_operand:SI 4 "register_operand" "=2,r")
	   (plus:SI (mult:SI (match_dup 0)
			     (match_dup 1))
		    (match_dup 2)))])]
  "REG_P (operands[2]) && REG_P (operands[4])
   && REGNO (operands[4]) == REGNO (operands[2])"
  [(set (match_dup 4) (plus:SI (mult:SI (match_dup 0) (match_dup 1))
				(match_dup 2)))
   (set (match_dup 5) (match_dup 3))]
  "operands[5] = gen_rtx (MEM, GET_MODE (operands[3]), operands[4]);")

(define_split
  [(parallel
    [(set (mem (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			(match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	   (match_operand 2 "register_operand" "r,r,r"))
     (set (match_operand:SI 3 "register_operand" "=0,r,r")
	   (plus:SI (match_dup 0)
		   (match_dup 1)))])]
  "(rtx_equal_p (operands[3], operands[0])
    || rtx_equal_p (operands[3], operands[1]))"
  [(set (match_dup 3) (plus:SI (match_dup 0) (match_dup 1)))
   (set (match_dup 5) (match_dup 2))]
  "operands[5] = gen_rtx (MEM, GET_MODE (operands[2]), operands[3]);")

(define_split
  [(parallel
    [(set (mem:SI (plus:SI
		    (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			     (match_operand:SI 1 "const_int_operand" "n,n"))
		    (match_operand:SI 2 "register_operand" "r,r")))
	   (const_int 0))
     (set (match_operand:SI 3 "register_operand" "=2,r")
	   (plus:SI (mult:SI (match_dup 0)
			     (match_dup 1))
		    (match_dup 2)))])]
  "REG_P (operands[2]) && REG_P (operands[3])
   && REGNO (operands[3]) == REGNO (operands[2])"
  [(set (match_dup 3) (plus:SI (mult:SI (match_dup 0) (match_dup 1))
				(match_dup 2)))
   (set (mem:SI (match_dup 3)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (mem:HI (plus:SI
		    (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			     (match_operand:SI 1 "const_int_operand" "n,n"))
		    (match_operand:SI 2 "register_operand" "r,r")))
	   (const_int 0))
     (set (match_operand:SI 3 "register_operand" "=2,r")
	   (plus:SI (mult:SI (match_dup 0)
			     (match_dup 1))
		    (match_dup 2)))])]
  "REG_P (operands[2]) && REG_P (operands[3])
   && REGNO (operands[3]) == REGNO (operands[2])"
  [(set (match_dup 3) (plus:SI (mult:SI (match_dup 0) (match_dup 1))
				(match_dup 2)))
   (set (mem:HI (match_dup 3)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (mem:QI (plus:SI
		    (mult:SI (match_operand:SI 0 "register_operand" "r,r")
			     (match_operand:SI 1 "const_int_operand" "n,n"))
		    (match_operand:SI 2 "register_operand" "r,r")))
	   (const_int 0))
     (set (match_operand:SI 3 "register_operand" "=2,r")
	   (plus:SI (mult:SI (match_dup 0)
			     (match_dup 1))
		    (match_dup 2)))])]
  "REG_P (operands[2]) && REG_P (operands[3])
   && REGNO (operands[3]) == REGNO (operands[2])"
  [(set (match_dup 3) (plus:SI (mult:SI (match_dup 0) (match_dup 1))
				(match_dup 2)))
   (set (mem:QI (match_dup 3)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (mem:SI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			   (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	   (const_int 0))
     (set (match_operand:SI 2 "register_operand" "=0,r,r")
	   (plus:SI (match_dup 0)
		    (match_dup 1)))])]
  "(rtx_equal_p (operands[0], operands[2])
    || rtx_equal_p (operands[2], operands[1]))"
  [(set (match_dup 2) (plus:SI (match_dup 0) (match_dup 1)))
   (set (mem:SI (match_dup 2)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (mem:HI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			   (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	   (const_int 0))
     (set (match_operand:SI 2 "register_operand" "=0,r,r")
	   (plus:SI (match_dup 0)
		    (match_dup 1)))])]
  "(rtx_equal_p (operands[0], operands[2])
    || rtx_equal_p (operands[2], operands[1]))"
  [(set (match_dup 2) (plus:SI (match_dup 0) (match_dup 1)))
   (set (mem:HI (match_dup 2)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (mem:QI (plus:SI (match_operand:SI 0 "bdap_operand" "%r,r,r")
			   (match_operand:SI 1 "bdap_operand" "r>Ri,r,>Ri")))
	   (const_int 0))
     (set (match_operand:SI 2 "register_operand" "=0,r,r")
	   (plus:SI (match_dup 0)
		    (match_dup 1)))])]
  "(rtx_equal_p (operands[0], operands[2])
    || rtx_equal_p (operands[2], operands[1]))"
  [(set (match_dup 2) (plus:SI (match_dup 0) (match_dup 1)))
   (set (mem:QI (match_dup 2)) (const_int 0))]
  "")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	  (match_operator
	    5 "extend_operator"
	    [(mem (plus:SI
		   (mult:SI (match_operand:SI 1 "register_operand" "r,r")
			    (match_operand:SI 2 "const_int_operand" "n,n"))
		   (match_operand:SI 3 "register_operand" "r,r")))]))
     (set (match_operand:SI 4 "register_operand" "=3,r")
	   (plus:SI (mult:SI (match_dup 1)
			     (match_dup 2))
		    (match_dup 3)))])]
  "REG_P (operands[3]) && REG_P (operands[4])
   && REGNO (operands[3]) == REGNO (operands[4])"
  [(set (match_dup 4) (plus:SI (mult:SI (match_dup 1) (match_dup 2))
				(match_dup 3)))
   (set (match_dup 0) (match_op_dup 5 [(match_dup 6)]))]
  "operands[6] = gen_rtx (MEM, GET_MODE (XEXP (operands[5],0)),
			   operands[4]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	  (match_operator
	    4 "extend_operator"
	    [(mem (plus:SI
		   (match_operand:SI 1 "bdap_operand" "%r,r,r")
		   (match_operand:SI 2 "bdap_operand" "r>Ri,r,>Ri")))]))
     (set (match_operand:SI 3 "register_operand" "=1,r,r")
	   (plus:SI (match_dup 1)
		    (match_dup 2)))])]
  "(rtx_equal_p (operands[1], operands[3])
    || rtx_equal_p (operands[2], operands[3]))"
  [(set (match_dup 3) (plus:SI (match_dup 1) (match_dup 2)))
   (set (match_dup 0) (match_op_dup 4 [(match_dup 5)]))]
  "operands[5] = gen_rtx (MEM, GET_MODE (XEXP (operands[4], 0)),
			  operands[3]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	  (match_operator
	    5 "orthogonal_operator"
	    [(match_operand 1 "register_operand" "0,0,0")
	     (mem (plus:SI
		   (match_operand:SI 2 "bdap_operand" "%r,r,r")
		   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))]))
     (set (match_operand:SI 4 "register_operand" "=2,r,r")
	   (plus:SI (match_dup 2)
		    (match_dup 3)))])]
  "(rtx_equal_p (operands[4], operands[2])
    || rtx_equal_p (operands[4], operands[3]))"
  [(set (match_dup 4) (plus:SI (match_dup 2) (match_dup 3)))
   (set (match_dup 0) (match_op_dup 5 [(match_dup 1) (match_dup 6)]))]
  "operands[6] = gen_rtx (MEM, GET_MODE (operands[0]), operands[4]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	  (match_operator
	    6 "orthogonal_operator"
	    [(match_operand 1 "register_operand" "0,0")
	     (mem (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))]))
     (set (match_operand:SI 5 "register_operand" "=4,r")
	   (plus:SI (mult:SI (match_dup 2)
			     (match_dup 3))
		   (match_dup 4)))])]
  "REG_P (operands[4]) && REG_P (operands[5])
   && REGNO (operands[5]) == REGNO (operands[4])"
  [(set (match_dup 5) (plus:SI (mult:SI (match_dup 2) (match_dup 3))
				(match_dup 4)))
   (set (match_dup 0) (match_op_dup 6 [(match_dup 1) (match_dup 7)]))]
  "operands[7] = gen_rtx (MEM, GET_MODE (operands[0]), operands[5]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	  (match_operator
	    6 "commutative_orth_op"
	    [(mem (plus:SI
		   (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			    (match_operand:SI 3 "const_int_operand" "n,n"))
		   (match_operand:SI 4 "register_operand" "r,r")))
	     (match_operand 1 "register_operand" "0,0")]))
     (set (match_operand:SI 5 "register_operand" "=4,r")
	   (plus:SI (mult:SI (match_dup 2)
			     (match_dup 3))
		    (match_dup 4)))])]
  "REG_P (operands[4]) && REG_P (operands[5])
   && REGNO (operands[5]) == REGNO (operands[4])"
  [(set (match_dup 5) (plus:SI (mult:SI (match_dup 2) (match_dup 3))
			       (match_dup 4)))
   (set (match_dup 0) (match_op_dup 6 [(match_dup 7) (match_dup 1)]))]
  "operands[7] = gen_rtx (MEM, GET_MODE (operands[0]), operands[5]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	  (match_operator
	    5 "commutative_orth_op"
	    [(mem (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			   (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))
	     (match_operand 1 "register_operand" "0,0,0")]))
     (set (match_operand:SI 4 "register_operand" "=2,r,r")
	  (plus:SI (match_dup 2)
		    (match_dup 3)))])]
  "(rtx_equal_p (operands[4], operands[2])
    || rtx_equal_p (operands[4], operands[3]))"
  [(set (match_dup 4) (plus:SI (match_dup 2) (match_dup 3)))
   (set (match_dup 0) (match_op_dup 5 [(match_dup 6) (match_dup 1)]))]
  "operands[6] = gen_rtx (MEM, GET_MODE (operands[0]), operands[4]);")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	  (match_operator
	    6 "operand_extend_operator"
	    [(match_operand 1 "register_operand" "0,0")
	     (match_operator
	      7 "extend_operator"
	      [(mem (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])]))
     (set (match_operand:SI 5 "register_operand" "=4,r")
	   (plus:SI (mult:SI (match_dup 2)
			     (match_dup 3))
		    (match_dup 4)))])]
  "REG_P (operands[4]) && REG_P (operands[5])
   && REGNO (operands[5]) == REGNO (operands[4])"
  [(set (match_dup 5) (plus:SI (mult:SI (match_dup 2) (match_dup 3))
			       (match_dup 4)))
   (set (match_dup 0) (match_op_dup 6 [(match_dup 1) (match_dup 8)]))]
  "operands[8] = gen_rtx (GET_CODE (operands[7]), GET_MODE (operands[7]),
			  gen_rtx (MEM, GET_MODE (XEXP (operands[7], 0)),
				   operands[5]));")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	  (match_operator
	    5 "operand_extend_operator"
	    [(match_operand 1 "register_operand" "0,0,0")
	     (match_operator
	      6 "extend_operator"
	      [(mem (plus:SI (match_operand:SI 2 "bdap_operand" "%r,r,r")
			     (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")
			     ))])]))
     (set (match_operand:SI 4 "register_operand" "=2,r,r")
	   (plus:SI (match_dup 2)
		    (match_dup 3)))])]
  "(rtx_equal_p (operands[4], operands[2])
    || rtx_equal_p (operands[4], operands[3]))"
  [(set (match_dup 4) (plus:SI (match_dup 2) (match_dup 3)))
   (set (match_dup 0) (match_op_dup 5 [(match_dup 1) (match_dup 7)]))]
  "operands[7] = gen_rtx (GET_CODE (operands[6]), GET_MODE (operands[6]),
			   gen_rtx (MEM, GET_MODE (XEXP (operands[6], 0)),
				    operands[4]));")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r")
	  (match_operator
	    7 "plus_or_bound_operator"
	    [(match_operator
	      6 "extend_operator"
	      [(mem (plus:SI
		     (mult:SI (match_operand:SI 2 "register_operand" "r,r")
			      (match_operand:SI 3 "const_int_operand" "n,n"))
		     (match_operand:SI 4 "register_operand" "r,r")))])
	     (match_operand 1 "register_operand" "0,0")]))
     (set (match_operand:SI 5 "register_operand" "=4,r")
	   (plus:SI (mult:SI (match_dup 2)
			     (match_dup 3))
		    (match_dup 4)))])]
  "REG_P (operands[4]) && REG_P (operands[5])
   && REGNO (operands[5]) == REGNO (operands[4])"
  [(set (match_dup 5) (plus:SI (mult:SI (match_dup 2) (match_dup 3))
			       (match_dup 4)))
   (set (match_dup 0) (match_op_dup 6 [(match_dup 8) (match_dup 1)]))]
  "operands[8] = gen_rtx (GET_CODE (operands[6]), GET_MODE (operands[6]),
			  gen_rtx (MEM, GET_MODE (XEXP (operands[6], 0)),
				   operands[5]));")

(define_split
  [(parallel
    [(set (match_operand 0 "register_operand" "=r,r,r")
	  (match_operator
	    6 "plus_or_bound_operator"
	    [(match_operator
	      5 "extend_operator"
	     [(mem (plus:SI
		    (match_operand:SI 2 "bdap_operand" "%r,r,r")
		    (match_operand:SI 3 "bdap_operand" "r>Ri,r,>Ri")))])
	     (match_operand 1 "register_operand" "0,0,0")]))
     (set (match_operand:SI 4 "register_operand" "=2,r,r")
	   (plus:SI (match_dup 2)
		    (match_dup 3)))])]
  "(rtx_equal_p (operands[4], operands[2])
    || rtx_equal_p (operands[4], operands[3]))"
  [(set (match_dup 4) (plus:SI (match_dup 2) (match_dup 3)))
   (set (match_dup 0) (match_op_dup 6 [(match_dup 7) (match_dup 1)]))]
  "operands[7] = gen_rtx (GET_CODE (operands[5]), GET_MODE (operands[5]),
			  gen_rtx (MEM, GET_MODE (XEXP (operands[5], 0)),
				   operands[4]));")

;;
;; Lots of opportunities to split up addressing prefixes, so we can
;; fill a delay slot.  Never split if we lose something, though.
;;
;;   move [indirect_ref],rx  indirect ref = {const, [r+], [r]}
;; costs as much as
;;   move indirect_ref,rx \; move [rx],rx
;; Just do not allow indirect_ref = register.
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operand 1 "indirect_operand" "m"))]
  "REG_P (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && (GET_CODE (XEXP (operands[1], 0)) == MEM
       || CONSTANT_P (XEXP (operands[1], 0)))"
  [(set (match_dup 2) (match_dup 4))
   (set (match_dup 0) (match_dup 3))]
  "operands[2] = gen_rtx (REG, Pmode, REGNO (operands[0]));
   operands[3] = gen_rtx (MEM, GET_MODE (operands[0]), operands[2]);
   operands[4] = XEXP (operands[1], 0);")

; movs, movu as above
(define_split
  [(set (match_operand 0 "register_operand" "=r")
	(match_operator
	 4 "extend_operator"
	 [(match_operand 1 "indirect_operand" "m")]))]
  "REG_P (operands[0])
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD
   && (GET_CODE (XEXP (operands[1], 0)) == MEM
       || CONSTANT_P (XEXP (operands[1], 0)))"
  [(set (match_dup 2) (match_dup 5))
   (set (match_dup 0) (match_op_dup 4 [(match_dup 3)]))]
  "operands[2] = gen_rtx (REG, Pmode, REGNO (operands[0]));
   operands[3] = gen_rtx (MEM, GET_MODE (XEXP (operands[4], 0)), operands[2]);
   operands[5] = XEXP (operands[1], 0);")

;;
;; Various peephole optimizations.
;;
;; WATCH OUT when you exchange one instruction for another, the
;; condition codes setting must be the same, or you have to CC_INIT (or
;; whatever you think is needed) in the pattern, before you emit the
;; assembly text.  This is best done here and not in
;; "notice_update_cc ()", to reduce the number of places to modify.
;;
;; Do not add patterns that you do not know will be matched.
;; Please also add a self-contained test-case.
;;

;;
;; We have trouble with and:s that could be transformed into
;; zero_extend:s and reloading them.  Maybe something is broken in gcc?
;; Or it could just be that bitfield insn expanding is a bit
;; unoptimized when not having extzv insns.
;;
(define_peephole
  [(set (match_operand 0 "register_operand" "=r")
	(ashiftrt:SI (match_dup 0)
		     (match_operand:SI 1 "const_int_operand" "n")))
   (set (match_dup 0)
	(and:SI (match_dup 0)
		(match_operand 2 "const_int_operand" "n")))]
  "INTVAL (operands[2]) > 31 && INTVAL (operands[2]) < 255
   && INTVAL (operands[1]) > 23"
; The m flag should be ignored, because this will be a *byte* "and" operation.
  "*
{
  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"lsrq %1,%0\;and.b %2,%0\";
}")

(define_peephole
  [(set (match_operand 0 "register_operand" "=r")
	(ashiftrt:SI (match_dup 0)
		     (match_operand:SI 1 "const_int_operand" "n")))
   (set (match_dup 0)
	(and:SI (match_dup 0)
		(match_operand 2 "const_int_operand" "n")))]
  "INTVAL (operands[2]) > 31 && INTVAL (operands[2]) < 65535
   && INTVAL (operands[2]) != 255
   && INTVAL (operands[1]) > 15"
; The m flag should be ignored, because this will be a *word* "and" operation.
  "*
{
  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"lsrq %1,%0\;and.w %2,%0\";
}")

(define_peephole
  [(set (match_operand 0 "register_operand" "=r")
	(lshiftrt:SI (match_dup 0)
		     (match_operand:SI 1 "const_int_operand" "n")))
   (set (match_dup 0)
	(and:SI (match_dup 0)
		(match_operand 2 "const_int_operand" "n")))]
  "INTVAL (operands[2]) > 31 && INTVAL (operands[2]) < 255
   && INTVAL (operands[1]) > 23"
; The m flag should be ignored, because this will be a *byte* "and" operation.
  "*
{
  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"lsrq %1,%0\;and.b %2,%0\";
}")

(define_peephole
  [(set (match_operand 0 "register_operand" "=r")
	(lshiftrt:SI (match_dup 0)
		     (match_operand:SI 1 "const_int_operand" "n")))
   (set (match_dup 0)
	(and:SI (match_dup 0)
		(match_operand 2 "const_int_operand" "n")))]
  "INTVAL (operands[2]) > 31 && INTVAL (operands[2]) < 65535
   && INTVAL (operands[2]) != 255
   && INTVAL (operands[1]) > 15"
; The m flag should be ignored, because this will be a *word* "and" operation.
  "*
{
  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"lsrq %1,%0\;and.w %2,%0\";
}")

;;
;; Change add.d n,rx move [rx],ry into move [rx=rx+n],ry
;; This will reduce the size of the assembler code for n = [-128..127],
;; and speed up accordingly.
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "const_int_operand" "n")))
   (set (match_operand 3 "register_operand" "=r")
	(mem (match_dup 0)))]
  "GET_MODE (operands[3]) != DImode
    && REGNO (operands[3]) != REGNO (operands[0])
    && (BASE_P (operands[1]) || BASE_P (operands[2]))
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
    && (INTVAL (operands[2]) >= -128 && INTVAL (operands[2]) < 128)"
  "move.%s3 [%0=%1%S2],%3")

;;
;; Vice versa.
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "const_int_operand" "n")))
   (set (mem (match_dup 0))
	(match_operand 3 "register_operand" "=r"))]
  "GET_MODE (operands[3]) != DImode
    && REGNO (operands[3]) != REGNO (operands[0])
    && (BASE_P (operands[1]) || BASE_P (operands[2]))
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
    && (INTVAL (operands[2]) >= -128 && INTVAL (operands[2]) < 128)"
  "move.%s3 %3,[%0=%1%S2]"
  [(set_attr "cc" "none")])

;;
;; Change add.d n,rx op.d [rx],ry into op.d [rx=rx+n],ry
;;  Saves when n = [-128..127], harmless otherwise.
;; We don't care about other valid combinations for sideeffect modes, they
;; probably will not save the time they take typing in, at this stage.
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(plus:SI (match_operand:SI 1 "register_operand" "0")
		 (match_operand:SI 2 "const_int_operand" "n")))
   (set (match_operand 3 "register_operand" "=r")
	(match_operator 4 "orthogonal_operator"
			   [(match_dup 3)
			    (mem (match_dup 0))]))]
  "GET_MODE (operands[3]) != DImode
    && REGNO (operands[0]) != REGNO (operands[3])
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'J')
    && !CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'N')
    && (INTVAL (operands[2]) >= -128 && INTVAL (operands[2]) < 128)"
  "%x4.%s0 [%3=%1%S2],%0")

;;
;; Sometimes, for some reason the pattern
;;  move x,rx
;;  add y,rx
;;  move [rx],rz
;; will occur.	Solve this, and likewise for operand.
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operand:SI 1 "bdap_biap_operand" "r,>Ri,r,>Ri"))
   (set (match_dup 0)
	(plus:SI (match_operand:SI 2 "bdap_biap_operand" "0,0,r>Ri,r")
		 (match_operand:SI 3 "bdap_biap_operand" "r>Ri,r,0,0")))
   (set (match_operand 4 "register_operand" "=r,r,r,r")
	(mem (match_dup 0)))]
  "(rtx_equal_p (operands[2], operands[0])
    || rtx_equal_p (operands[3], operands[0]))
   && side_effect_mode_ok (PLUS, operands, 0,
			   (REG_S_P (operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   (!REG_S_P(operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   -1, 4)"
  "@
   move.%s4 [%0=%1%S3],%4
   move.%s4 [%0=%3%S1],%4
   move.%s4 [%0=%1%S2],%4
   move.%s4 [%0=%2%S1],%4")

;;
;; As above but to memory.
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operand:SI 1 "bdap_biap_operand" "r,>Ri,r,>Ri"))
   (set (match_dup 0)
	(plus:SI (match_operand:SI 2 "bdap_biap_operand" "0,0,r>Ri,r")
		 (match_operand:SI 3 "bdap_biap_operand" "r>Ri,r,0,0")))
   (set (mem (match_dup 0))
	(match_operand 4 "register_operand" "=r,r,r,r"))]
  "(rtx_equal_p (operands[2], operands[0])
    || rtx_equal_p (operands[3], operands[0]))
   && side_effect_mode_ok (PLUS, operands, 0,
			   (REG_S_P (operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   (!REG_S_P (operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   -1, 4)"
  "@
   move.%s4 %4,[%0=%1%S3]
   move.%s4 %4,[%0=%3%S1]
   move.%s4 %4,[%0=%1%S2]
   move.%s4 %4,[%0=%2%S1]"
  [(set_attr "cc" "none")])

;;
;; Op as above.
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operand:SI 1 "bdap_biap_operand" "r,>Ri,r,>Ri"))
   (set (match_dup 0)
	(plus:SI (match_operand:SI 2 "bdap_biap_operand" "0,0,r>Ri,r")
		 (match_operand:SI 3 "bdap_biap_operand" "r>Ri,r,0,0")))
   (set (match_operand 4 "register_operand" "=r,r,r,r")
	(match_operator 5 "orthogonal_operator"
			[(match_dup 3)
			 (mem (match_dup 0))]))]
  "(rtx_equal_p (operands[2], operands[0])
    || rtx_equal_p (operands[3], operands[0]))
    && side_effect_mode_ok (PLUS, operands, 0,
			    (REG_S_P (operands[1])
			     ? 1
			     : (rtx_equal_p (operands[2], operands[0])
				? 3 : 2)),
			    (!REG_S_P (operands[1])
			     ? 1
			     : (rtx_equal_p (operands[2], operands[0])
				? 3 : 2)),
			    -1, 4)"
  "@
   %x5.%s4 [%0=%1%S3],%4
   %x5.%s4 [%0=%3%S1],%4
   %x5.%s4 [%0=%1%S2],%4
   %x5.%s4 [%0=%2%S1],%4")

;; Same, but with swapped operands (and commutative operation).
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r,r,r,r")
	(match_operand:SI 1 "bdap_biap_operand" "r,>Ri,r,>Ri"))
   (set (match_dup 0)
	(plus:SI (match_operand:SI 2 "bdap_biap_operand" "0,0,r>Ri,r")
		 (match_operand:SI 3 "bdap_biap_operand" "r>Ri,r,0,0")))
   (set (match_operand 4 "register_operand" "=r,r,r,r")
	(match_operator 5 "commutative_orth_op"
			[(mem (match_dup 0))
			 (match_dup 3)]))]
  "(rtx_equal_p (operands[2], operands[0])
    || rtx_equal_p (operands[3], operands[0]))
   && side_effect_mode_ok (PLUS, operands, 0,
			   (REG_S_P (operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   (!REG_S_P (operands[1])
			    ? 1
			    : (rtx_equal_p (operands[2], operands[0])
			       ? 3 : 2)),
			   -1, 4)"
  "@
   %x5.%s4 [%0=%1%S3],%4
   %x5.%s4 [%0=%3%S1],%4
   %x5.%s4 [%0=%1%S2],%4
   %x5.%s4 [%0=%2%S1],%4")

;;
;; Another braindamage: move rx,ry \; move [ry],ry
;;
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(match_operand:SI 1 "register_operand" "r"))
   (set (match_operand 2 "register_operand" "=r")
	(mem (match_dup 0)))]
  "REGNO (operands[0]) == REGNO (operands[2])
   && GET_MODE_SIZE (GET_MODE (operands[2])) <= UNITS_PER_WORD"
  "move.%s2 [%1],%0"
  [(set_attr "slottable" "yes")])

;; And a simple variant with extended operand.
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(match_operand:SI 1 "register_operand" "r"))
   (set (match_operand 2 "register_operand" "=r")
	(match_operator 3 "extend_operator" [(mem (match_dup 0))]))]
  "REGNO (operands[0]) == REGNO (operands[2])
   && GET_MODE_SIZE (GET_MODE (operands[2])) <= UNITS_PER_WORD"
  "mov%e3.%m3 [%1],%0"
  [(set_attr "slottable" "yes")])

;; Regression-controlled peepholes (with saved testcases).

;; peep-1:
;;   move.d [r10+16],r9 \; and.d r12,r9 => and.d [r10+16],r12,r9
;;  This seems to be the result of a quirk in register allocation
;; missing the three-operand cases when having different predicates (and
;; maybe that it matters that it is a commutative operation).
;;  This pattern helps that situation, but there's still the increased
;; register pressure.
;; (match HI and QI as well, and bdap_biap too, and plus and bound).
;;  Note that adding the noncommutative variant did not show any matches
;; in ipps and cc1.
(define_peephole
  [(set (match_operand 0 "register_operand" "=r,r,r,r")
	(mem (plus:SI
	      (match_operand:SI 1 "bdap_biap_operand" "r,r>Ri,r,r>Ri")
	      (match_operand:SI 2 "bdap_biap_operand" "r>Ri,r,r>Ri,r"))))
   (set (match_dup 0)
	(match_operator 5 "commutative_orth_op"
			[(match_operand 3 "register_operand" "0,0,r,r")
			 (match_operand 4 "register_operand" "r,r,0,0")]))]
  "(rtx_equal_p (operands[3], operands[0])
    || rtx_equal_p (operands[4], operands[0]))
   && !rtx_equal_p (operands[3], operands[4])
   && (REG_S_P (operands[1]) || REG_S_P (operands[2]))
   && GET_MODE_SIZE (GET_MODE (operands[0])) <= UNITS_PER_WORD"
  "@
   %x5.%s0 [%1%S2],%4,%0
   %x5.%s0 [%2%S1],%4,%0
   %x5.%s0 [%1%S2],%3,%0
   %x5.%s0 [%2%S1],%3,%0")

;; I cannot tell Gcc (2.1) how to reload an instruction that looks like
;; and.b mem_byte,const,reg_32
;; where reg is the destination of the "three-address" code optimally.
;;  It should be:
;; movu.b mem_byte,reg_32
;; and.b const,reg_32
;;  but is turns into:
;; move.b mem_byte,reg_32
;; and.d const,reg_32
;;  Fix it.
;; (Still in 2.7.2)
;;
;; peep-2
(define_peephole
  [(set (match_operand:SI 0 "register_operand" "=r")
	(match_operand:SI 1 "nonimmediate_operand" "rm"))
   (set (match_operand:SI 2 "register_operand" "=0")
	(and:SI (match_dup 0)
		(match_operand:SI 3 "const_int_operand" "n")))]
; Just bail out when volatile mems are involved.
  "REGNO (operands[2]) == REGNO (operands[0])
   && INTVAL (operands[3]) <= 65535 && INTVAL (operands[3]) >= 0
   && !CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'I')
   && !(GET_CODE (operands[1]) == MEM && MEM_VOLATILE_P (operands[1]))"
  "*
{
  if (CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'O'))
    return \"movu.%z3 %1,%0\;andq %b3,%0\";

  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"movu.%z3 %1,%0\;and.%z3 %3,%0\";
}")

;; peep-3
(define_peephole
  [(set (match_operand 0 "register_operand" "=r")
	(match_operand 1 "nonimmediate_operand" "rm"))
   (set (match_operand:SI 2 "register_operand" "=r")
	(and:SI (subreg:SI (match_dup 0) 0)
		(match_operand 3 "const_int_operand" "n")))]
; Just bail out when volatile mems are involved.
  "REGNO (operands[0]) == REGNO (operands[2])
   && INTVAL (operands[3]) > 0 &&  INTVAL (operands[3]) <= 65535
   && !CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'I')
   && !(GET_CODE (operands[1]) == MEM && MEM_VOLATILE_P (operands[1]))"
  "*
{
  if (CONST_OK_FOR_LETTER_P (INTVAL (operands[3]), 'O'))
    return \"movu.%z3 %1,%0\;andq %b3,%0\";

  cc_status.flags |= CC_NOT_NEGATIVE;

  return \"movu.%z3 %1,%0\;and.%z3 %3,%0\";
}")

;; Local variables:
;; mode:emacs-lisp
;; comment-start: ";; "
;; eval: (set-syntax-table (copy-sequence (syntax-table)))
;; eval: (modify-syntax-entry ?[ "(]")
;; eval: (modify-syntax-entry ?] ")[")
;; eval: (modify-syntax-entry ?{ "(}")
;; eval: (modify-syntax-entry ?} "){")
;; eval: (setq indent-tabs-mode t)
;; End:
